"""
This script contains the main SLISE functions, and classes.

The library can both be used "sk-learn" style with `SliseRegression(...).fit(X, y)`
and `SliseExplanation(...).explain(index)`, or in a more functional style with
`regression(...)` and `explain(...)`.
"""

from __future__ import annotations

from typing import Callable, List, Optional, Sequence, Tuple, Union
from warnings import warn

import numpy as np
from matplotlib.pyplot import Figure
from scipy.special import expit as sigmoid

from slise.data import (
    DataScaling,
    add_constant_columns,
    add_intercept_column,
    normalise_robust,
    remove_constant_columns,
)
from slise.initialisation import initialise_candidates, initialise_fixed
from slise.optimisation import (
    check_threading_layer,
    graduated_optimisation,
    loss_sharp,
    set_threads,
)
from slise.plot import plot_2d, plot_dist, plot_dist_single, plot_image, print_slise
from slise.utils import SliseWarning, limited_logit, mat_mul_inter


def regression(
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
    intercept: bool = True,
    normalise: bool = False,
    init: Union[None, np.ndarray, Tuple[np.ndarray, float]] = None,
    initialisation: Callable[
        [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float]
    ] = initialise_candidates,
    beta_max: float = 20,
    max_approx: float = 1.15,
    max_iterations: int = 300,
    debug: bool = False,
    num_threads: int = 1,
) -> SliseRegression:
    """Use SLISE for robust regression

    In robust regression we fit regression models that can handle data that
    contains outliers. SLISE accomplishes this by fitting a model such that
    the largest possible subset of the data items have an error less than a
    given value. All items with an error larger than that are considered
    potential outliers and do not affect the resulting model.

    It is highly recommended that you normalise the data, either before using SLISE or by setting normalise = TRUE.
    This is a wrapper around [slise.slise.SliseRegression][] that is equivalent to `SliseRegression(epsilon, **kwargs).fit(X, Y)`

    Args:
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        epsilon (float): Error tolerance.
        lambda1 (float, optional): L1 regularisation strength. Defaults to 0.
        lambda2 (float, optional): L2 regularisation strength. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
        intercept (bool, optional): Add an intercept term. Defaults to True.
        normalise (bool, optional): Should X aclasses not be scaled). Defaults to False.
        init (Union[None, np.ndarray, Tuple[np.ndarray, float]], optional): Use this alpha (and beta) value instead of the initialisation function. Defaults to None.
        initialisation (Callable[ [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float] ], optional): Function that takes `(X, Y, epsilon, weight)` and gives an initial values for alpha and beta. Defaults to initialise_candidates.
        beta_max (float, optional): The stopping sigmoid steepness. Defaults to 20.
        max_approx (float, optional): Approximation ratio when selecting the next beta. Defaults to 1.15.
        max_iterations (int, optional): Maximum number of OWL-QN iterations. Defaults to 300.
        debug (bool, optional): Print debug statements each graduated optimisation step. Defaults to False.
        num_threads (int, optional): The number of numba threads. Set to -1 to use numba defaults. Values >1 sometimes cause unexpectedly large overhead on some CPUs. Defaults to 1.

    Returns:
        SliseRegression: Object containing the regression result.
    """
    return SliseRegression(
        epsilon=epsilon,
        lambda1=lambda1,
        lambda2=lambda2,
        intercept=intercept,
        normalise=normalise,
        initialisation=initialisation,
        beta_max=beta_max,
        max_approx=max_approx,
        max_iterations=max_iterations,
        debug=debug,
        num_threads=num_threads,
    ).fit(X=X, Y=Y, weight=weight, init=init)


def explain(
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    x: Union[np.ndarray, int],
    y: Union[float, None] = None,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
    normalise: bool = False,
    logit: bool = False,
    init: Union[None, np.ndarray, Tuple[np.ndarray, float]] = None,
    initialisation: Callable[
        [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float]
    ] = initialise_candidates,
    beta_max: float = 20,
    max_approx: float = 1.15,
    max_iterations: int = 300,
    debug: bool = False,
    num_threads: int = 1,
) -> SliseExplainer:
    """Use SLISE for explaining outcomes from black box models.

    SLISE can also be used to provide local model-agnostic explanations for
    outcomes from black box models. To do this we replace the ground truth
    response vector with the predictions from the complex model. Furthermore, we
    force the model to fit a selected item (making the explanation local). This
    gives us a local approximation of the complex model with a simpler linear
    model. In contrast to other methods SLISE creates explanations using real
    data (not some discretised and randomly sampled data) so we can be sure that
    all inputs are valid (i.e. in the correct data manifold, and follows the
    constraints used to generate the data, e.g., the laws of physics).

    It is highly recommended that you normalise the data, either before using SLISE or by setting normalise = TRUE.
    This is a wrapper around [slise.slise.SliseExplainer][] that is equivalent to `SliseExplainer(X, Y, epsilon, **kwargs).explain(x, y)`

    Args:
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Vector of predictions.
        epsilon (float): Error tolerance.
        x (Union[np.ndarray, int]): The data item to explain, or an index to get the item from self.X
        y (Union[float, None], optional): The outcome to explain. If x is an index then this should be None (y is taken from self.Y). Defaults to None.
        lambda1 (float, optional): L1 regularisation strength. Defaults to 0.
        lambda2 (float, optional): L2 regularisation strength. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
        normalise (bool, optional): Should X and Y be normalised (note that epsilon will not be scaled). Defaults to False.
        logit (bool, optional): Do a logit transformation on the Y vector, this is recommended only if Y consists of probabilities. Defaults to False.
        init (Union[None, np.ndarray, Tuple[np.ndarray, float]], optional): Use this alpha (and beta) value instead of the initialisation function. Defaults to None.
        initialisation (Callable[ [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float] ], optional): Function that takes `(X, Y, epsilon, weight)` and gives an initial values for alpha and beta. Defaults to initialise_candidates.
        beta_max (float, optional): The final sigmoid steepness. Defaults to 20.
        max_approx (float, optional): Approximation ratio when selecting the next beta. Defaults to 1.15.
        max_iterations (int, optional): Maximum number of OWL-QN iterations. Defaults to 300.
        debug (bool, optional): Print debug statements each graduated optimisation step. Defaults to False.
        num_threads (int, optional): The number of numba threads. Set to -1 to use numba defaults. Values >1 sometimes cause unexpectedly large overhead on some CPUs. Defaults to 1.

    Returns:
        SliseExplainer: Object containing the explanation.
    """
    return SliseExplainer(
        X=X,
        Y=Y,
        epsilon=epsilon,
        lambda1=lambda1,
        lambda2=lambda2,
        normalise=normalise,
        logit=logit,
        initialisation=initialisation,
        beta_max=beta_max,
        max_approx=max_approx,
        max_iterations=max_iterations,
        debug=debug,
        num_threads=num_threads,
    ).explain(x=x, y=y, weight=weight, init=init)


class SliseRegression:
    """
    Class for holding the result from using SLISE for regression.
    Can also be used sklearn-style to do regression.
    """

    def __init__(
        self,
        epsilon: float,
        lambda1: float = 0,
        lambda2: float = 0,
        intercept: bool = True,
        normalise: bool = False,
        initialisation: Callable[
            [np.ndarray, np.ndarray, float, Optional[np.ndarray]],
            Tuple[np.ndarray, float],
        ] = initialise_candidates,
        beta_max: float = 20,
        max_approx: float = 1.15,
        max_iterations: int = 300,
        debug: bool = False,
        num_threads: int = 1,
    ):
        """Use SLISE for robust regression.

        In robust regression we fit regression models that can handle data that
        contains outliers. SLISE accomplishes this by fitting a model such that
        the largest possible subset of the data items have an error less than a
        given value. All items with an error larger than that are considered
        potential outliers and do not affect the resulting model.

        This constructor prepares the parameters, call `fit` to fit a robust regression to a dataset.
        It is highly recommended that you normalise the data, either before using SLISE or by setting normalise = TRUE.

        Args:
            epsilon (float): Error tolerance.
            lambda1 (float, optional): L1 regularisation strength. Defaults to 0.
            lambda2 (float, optional): L2 regularisation strength. Defaults to 0.
            intercept (bool, optional): Add an intercept term. Defaults to True.
            normalise (bool, optional): Should X and Y be normalised (note that epsilon will not be scaled). Defaults to False.
            initialisation (Callable[ [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float] ], optional): Function that takes `(X, Y, epsilon, weight)` and gives an initial values for alpha and beta. Defaults to initialise_candidates.
            beta_max (float, optional): The stopping sigmoid steepness. Defaults to 20.
            max_approx (float, optional): Approximation ratio when selecting the next beta. Defaults to 1.15.
            max_iterations (int, optional): Maximum number of OWL-QN iterations. Defaults to 300.
            debug (bool, optional): Print debug statements each graduated optimisation step. Defaults to False.
            num_threads (int, optional): The number of numba threads. Set to -1 to use numba defaults. Values >1 sometimes cause unexpectedly large overhead on some CPUs. Defaults to 1.
        """
        assert epsilon > 0.0, "`epsilon` must be positive!"
        assert lambda1 >= 0.0, "`lambda1` must not be negative!"
        assert lambda2 >= 0.0, "`lambda2` must not be negative!"
        assert beta_max > 0.0, "`beta_max` must be positive!"
        assert max_approx > 1.0, "`max_approx` must be larger than 1.0!"
        assert max_iterations > 0, "`max_iterations` must be positive!"
        self.epsilon = epsilon
        self.lambda1 = lambda1
        self.lambda2 = lambda2
        self.init_fn = initialisation
        self.beta_max = beta_max
        self.max_approx = max_approx
        self.max_iterations = max_iterations
        self.debug = debug
        self._intercept = intercept
        self._normalise = normalise
        self._scale = None
        self._X = None
        self._Y = None
        self._weight = None
        self._alpha = None
        self._coefficients = None
        self.num_threads = num_threads
        check_threading_layer()

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        weight: Optional[np.ndarray] = None,
        init: Union[None, np.ndarray, Tuple[np.ndarray, float]] = None,
    ) -> SliseRegression:
        """Robustly fit a linear regression to a dataset

        Args:
            X (np.ndarray): Data matrix.
            Y (np.ndarray): Response vector.
            weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
            init (Union[None, np.ndarray, Tuple[np.ndarray, float]], optional): Use this alpha (and beta) value instead of the initialisation function. Defaults to None.

        Returns:
            SliseRegression: `self` (containing the regression result).
        """
        X = np.array(X)
        Y = np.array(Y)
        if len(X.shape) == 1:
            X.shape = X.shape + (1,)
        assert X.shape[0] == Y.shape[0], "X and Y must have the same number of items!"
        if len(Y.shape) > 1:
            Y = Y.ravel()
            assert X.shape[0] == Y.shape[0], "Y cannot have multiple columns!"
        self._X = X
        self._Y = Y
        if weight is None:
            self._weight = None
        else:
            self._weight = np.array(weight)
            assert len(self._weight) == len(
                self._Y
            ), "Y and weight must have the same number of items!"
            assert np.all(self._weight >= 0.0), "Weights must not be negative!"
        # Preprocessing
        if self._normalise:
            X, x_cols = remove_constant_columns(X)
            if self._X.shape[1] == X.shape[1]:
                x_cols = None
            X, x_center, x_scale = normalise_robust(X)
            Y, y_center, y_scale = normalise_robust(Y)
            self._scale = DataScaling(x_center, x_scale, y_center, y_scale, x_cols)
        if self._intercept:
            X = add_intercept_column(X)
        # Initialisation
        threads = set_threads(self.num_threads)
        try:
            if init is None:
                alpha, beta = self.init_fn(X, Y, self.epsilon, self._weight)
            else:
                alpha, beta = initialise_fixed(init, X, Y, self.epsilon, self._weight)
            # Optimisation
            alpha = graduated_optimisation(
                alpha=alpha,
                X=X,
                Y=Y,
                epsilon=self.epsilon,
                beta=beta,
                lambda1=self.lambda1,
                lambda2=self.lambda2,
                weight=self._weight,
                beta_max=self.beta_max,
                max_approx=self.max_approx,
                max_iterations=self.max_iterations,
                debug=self.debug,
            )
        finally:
            set_threads(threads)
        self._alpha = alpha
        if self._normalise:
            alpha2 = self._scale.unscale_model(alpha)
            if not self._intercept:
                if np.abs(alpha2[0]) > 1e-8:
                    warn(
                        "Intercept introduced due to scaling, consider setting intercept=True (or normalise=False)",
                        SliseWarning,
                    )
                    self._intercept = True
                    self._alpha = np.concatenate(([0], alpha))
                else:
                    alpha2 = alpha2[1:]
            self._coefficients = alpha2
        else:
            self._coefficients = alpha
        return self

    def get_params(self, normalised: bool = False) -> np.ndarray:
        """Get the coefficients of the linear model.

        Args:
            normalised (bool, optional): If the data is normalised within SLISE, return a linear model ftting the normalised data. Defaults to False.

        Returns:
            np.ndarray: The coefficients of the linear model.
        """
        warn("Use `coefficients` instead of `get_params()`.", SliseWarning)
        return self._alpha if normalised else self._coefficients

    @property
    def coefficients(self) -> np.ndarray:
        """Get the coefficients of the linear model.

        Returns:
            np.ndarray: The coefficients of the linear model (the first scalar in the vector is the intercept).
        """
        if self._coefficients is None:
            warn("Fit the model before retrieving coefficients", SliseWarning)
        return self._coefficients

    def normalised(self, all_columns: bool = True) -> Optional[np.ndarray]:
        """Get coefficients for normalised data (if the data is normalised within SLISE).

        Args:
            all_columns (bool, optional): Add coefficients for constant columns. Defaults to True.

        Returns:
            Optional[np.ndarray]: The normalised coefficients or None.
        """
        if self._alpha is None:
            warn("Fit the model before retrieving coefficients", SliseWarning)
        if self._normalise:
            if all_columns:
                return add_constant_columns(self._alpha, self._scale.columns, True)
            else:
                return self._alpha
        else:
            return None

    @property
    def scaled_epsilon(self) -> float:
        """Espilon fitting unnormalised data (if the data is normalised).

        Returns:
            float: Scaled epsilon.
        """
        if self._normalise:
            return self.epsilon * self._scale.y_scale
        else:
            return self.epsilon

    def predict(self, X: Union[np.ndarray, None] = None) -> np.ndarray:
        """Use the fitted model to predict new responses.

        Args:
            X (Union[np.ndarray, None], optional): Data matrix to predict, or None for using the fitted dataset. Defaults to None.

        Returns:
            np.ndarray: Predicted response.
        """
        if X is None:
            return mat_mul_inter(self._X, self.coefficients)
        else:
            return mat_mul_inter(X, self.coefficients)

    def score(
        self, X: Union[np.ndarray, None] = None, Y: Union[np.ndarray, None] = None
    ) -> float:
        """Calculate the loss. Lower is better and it should usually be negative (unless the regularisation is very (too?) strong).

        Args:
            X (Union[np.ndarray, None], optional): Data matrix, or None for using the fitted dataset. Defaults to None.
            Y (Union[np.ndarray, None], optional): Response vector, or None for using the fitted dataset. Defaults to None.

        Returns:
            float: The loss.
        """
        if self._alpha is None:
            warn("Fit the model before calculating the score", SliseWarning)
        if X is None or Y is None:
            X = self._X
            Y = self._Y
        if self._normalise:
            X = self._scale.scale_x(X)
            Y = self._scale.scale_y(Y)
        return loss_sharp(
            self._alpha, X, Y, self.epsilon, self.lambda1, self.lambda2, self._weight
        )

    loss = score
    value = score

    def subset(
        self, X: Union[np.ndarray, None] = None, Y: Union[np.ndarray, None] = None
    ) -> np.ndarray:
        """Get the subset (of non-outliers) used for the robust regression model.

        Args:
            X (Union[np.ndarray, None], optional): Data matrix, or None for using the fitted dataset. Defaults to None.
            Y (Union[np.ndarray, None], optional): Response vector, or None for using the fitted dataset. Defaults to None.

        Returns:
            np.ndarray: The selected subset as a boolean mask.
        """
        if X is None or Y is None:
            X = self._X
            Y = self._Y
        Y2 = mat_mul_inter(X, self.coefficients)
        return (Y2 - Y) ** 2 < self.scaled_epsilon**2

    def print(
        self,
        variables: Union[List[str], None] = None,
        decimals: int = 3,
        num_var: int = 10,
    ):
        """Print the current robust regression result.

        Args:
            variables (Union[List[str], None], optional): Names of the variables/columns in X. Defaults to None.
            num_var (int, optional): Exclude zero weights if there are too many variables. Defaults to 10.
            decimals (int, optional): Precision to use for printing. Defaults to 3.
        """
        print_slise(
            self.coefficients,
            self._intercept,
            self.subset(),
            self.score(),
            self.scaled_epsilon,
            variables,
            "SLISE Regression",
            decimals,
            num_var,
            alpha=self.normalised(),
        )

    def plot_2d(
        self,
        title: str = "SLISE Regression",
        label_x: str = "x",
        label_y: str = "y",
        decimals: int = 3,
        fig: Union[Figure, None] = None,
    ) -> SliseRegression:
        """Plot the regression in a 2D scatter plot with a line for the regression model.

        Args:
            title (str, optional): Title of the plot. Defaults to "SLISE Regression".
            label_x (str, optional): X-axis label. Defaults to "x".
            label_y (str, optional): Y-axis label. Defaults to "y".
            decimals (int, optional): Number of decimals when writing numbers. Defaults to 3.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.

        Raises:
            SliseException: If the data has too many dimensions.
        """
        plot_2d(
            self._X,
            self._Y,
            self.coefficients,
            self.scaled_epsilon,
            None,
            None,
            False,
            title,
            label_x,
            label_y,
            decimals,
            fig,
        )

    def plot_dist(
        self,
        title: str = "SLISE Regression",
        variables: list = None,
        order: Union[None, int, Sequence[int]] = None,
        decimals: int = 3,
        fig: Union[Figure, None] = None,
    ) -> SliseExplainer:
        """Plot the regression with density distributions for the dataset and a barplot for the model.

        Args:
            title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
            variables (list, optional): Names for the variables. Defaults to None.
            order (Union[None, int, Sequence[int]], optional): Select variables (None: all, int: largest, selected). Defaults to all.
            decimals (int, optional): Number of decimals to write. Defaults to 3.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
        """
        plot_dist(
            X=self._X,
            Y=self._Y,
            model=self.coefficients,
            subset=self.subset(),
            alpha=self.normalised(),
            x=None,
            y=None,
            terms=None,
            norm_terms=None,
            title=title,
            variables=variables,
            order=order,
            decimals=decimals,
            fig=fig,
        )

    def plot_subset(
        self,
        title: str = "Response Distribution",
        decimals: int = 0,
        fig: Union[Figure, None] = None,
    ):
        """Plot a density distributions for response and the response of the subset

        Args:
            title (str, optional): Title of the plot. Defaults to "Response Distribution".
            decimals (int, optional): Number of decimals when writing the subset size. Defaults to 0.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
        """
        plot_dist_single(self._Y, self.subset(), None, title, decimals, fig)


class SliseExplainer:
    """
    Class for holding the result from using SLISE as an explainer.
    Can also be used sklearn-style to create explanations.
    """

    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        epsilon: float,
        lambda1: float = 0,
        lambda2: float = 0,
        logit: bool = False,
        normalise: bool = False,
        initialisation: Callable[
            [np.ndarray, np.ndarray, float, Optional[np.ndarray]],
            Tuple[np.ndarray, float],
        ] = initialise_candidates,
        beta_max: float = 20,
        max_approx: float = 1.15,
        max_iterations: int = 300,
        debug: bool = False,
        num_threads: int = 1,
    ):
        """Use SLISE for explaining outcomes from black box models.

        SLISE can also be used to provide local model-agnostic explanations for
        outcomes from black box models. To do this we replace the ground truth
        response vector with the predictions from the complex model.
        Furthermore, we force the model to fit a selected item (making the
        explanation local). This gives us a local approximation of the complex
        model with a simpler linear model. In contrast to other methods SLISE
        creates explanations using real data (not some discretised and randomly
        sampled data) so we can be sure that all inputs are valid (i.e. in the
        correct data manifold, and follows the constraints used to generate the
        data, e.g., the laws of physics).

        This prepares the dataset used for the explanations, call `explain` on this object to explain outcomes.
        It is highly recommended that you normalise the data, either before using SLISE or by setting normalise = TRUE.

        Args:
            X (np.ndarray): Data matrix.
            Y (np.ndarray): Vector of predictions.
            epsilon (float): Error tolerance.
            lambda1 (float, optional): L1 regularisation strength. Defaults to 0.
            lambda2 (float, optional): L2 regularisation strength. Defaults to 0.
            logit (bool, optional): Do a logit transformation on the Y vector, this is recommended opnly if Y consists of probabilities. Defaults to False.
            normalise (bool, optional): Should X and Y be normalised (note that epsilon will not be scaled). Defaults to False.
            initialisation (Callable[ [np.ndarray, np.ndarray, float, Optional[np.ndarray]], Tuple[np.ndarray, float] ], optional): Function that takes `(X, Y, epsilon, weight)` and gives an initial values for alpha and beta. Defaults to initialise_candidates.
            beta_max (float, optional): The final sigmoid steepness. Defaults to 20.
            max_approx (float, optional): Approximation ratio when selecting the next beta. Defaults to 1.15.
            max_iterations (int, optional): Maximum number of OWL-QN iterations. Defaults to 300.
            debug (bool, optional): Print debug statements each graduated optimisation step. Defaults to False.
            num_threads (int, optional): The number of numba threads. Set to -1 to use numba defaults. Values >1 sometimes cause unexpectedly large overhead on some CPUs. Defaults to 1.
        """
        assert epsilon > 0.0, "`epsilon` must be positive!"
        assert lambda1 >= 0.0, "`lambda1` must not be negative!"
        assert lambda2 >= 0.0, "`lambda2` must not be negative!"
        assert beta_max > 0.0, "`beta_max` must be positive!"
        assert max_approx > 1.0, "`max_approx` must be larger than 1.0!"
        assert max_iterations > 0, "`max_iterations` must be positive!"
        self.epsilon = epsilon
        self.lambda1 = lambda1
        self.lambda2 = lambda2
        self.init_fn = initialisation
        self.beta_max = beta_max
        self.max_approx = max_approx
        self.max_iterations = max_iterations
        self.debug = debug
        X = np.array(X)
        Y = np.array(Y)
        if len(X.shape) == 1:
            X.shape = X.shape + (1,)
        assert X.shape[0] == Y.shape[0], "X and Y must have the same number of items"
        if len(Y.shape) > 1:
            Y = Y.ravel()
            assert X.shape[0] == Y.shape[0], "Y cannot have multiple columns!"
        self._logit = logit
        self._normalise = normalise
        self._X = X
        self._Y = Y
        self._x = None
        self._y = None
        self._weight = None
        self._alpha = None
        self._coefficients = None
        # Preprocess data
        if logit:
            Y = limited_logit(Y)
        if normalise:
            X2, x_cols = remove_constant_columns(X)
            if X.shape[1] == X2.shape[1]:
                x_cols = None
            X, x_center, x_scale = normalise_robust(X2)
            if logit:
                (y_center, y_scale) = (0, 1)
            else:
                Y, y_center, y_scale = normalise_robust(Y)
            self._scale = DataScaling(x_center, x_scale, y_center, y_scale, x_cols)
        else:
            self._scale = None
        self._X2 = X
        self._Y2 = Y
        self.num_threads = num_threads
        check_threading_layer()

    def explain(
        self,
        x: Union[np.ndarray, int],
        y: Union[float, None] = None,
        weight: Optional[np.ndarray] = None,
        init: Union[None, np.ndarray, Tuple[np.ndarray, float]] = None,
    ) -> SliseExplainer:
        """Explain an outcome from a black box model

        Args:
            x (Union[np.ndarray, int]): Data item to explain, or an index to get the item from self.X
            y (Union[float, None], optional): Prediction to explain. If x is an index then this should be None (y is taken from self.Y). Defaults to None.
            weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
            init (Union[None, np.ndarray, Tuple[np.ndarray, float]], optional): Use this alpha (and beta) value instead of the initialisation function. Defaults to None.

        Returns:
            SliseExplainer: `self` (containing the explanation).
        """
        if weight is None:
            self._weight = None
        else:
            self._weight = np.array(weight)
            assert len(self._weight) == len(
                self._Y
            ), "Y and weight must have the same number of items!"
            assert np.all(self._weight >= 0.0), "Weights must not be negative!"
        if y is None:
            assert isinstance(x, int) and (
                0 <= x < self._Y.shape[0]
            ), "If y is None then x must be an integer index [0, len(Y)["
            self._y = self._Y[x]
            self._x = self._X[x, :]
            y = self._Y2[x]
            x = self._X2[x, :]
        else:
            x = np.atleast_1d(np.array(x))
            self._x = x
            self._y = y
            if self._logit:
                y = limited_logit(y)
            if self._normalise:
                x = self._scale.scale_x(x)
                y = self._scale.scale_y(y)
        X = self._X2 - x[None, :]
        Y = self._Y2 - y
        threads = set_threads(self.num_threads)
        try:
            if init is None:
                alpha, beta = self.init_fn(X, Y, self.epsilon, self._weight)
            else:
                alpha, beta = initialise_fixed(init, X, Y, self.epsilon, self._weight)
            alpha = graduated_optimisation(
                alpha=alpha,
                X=X,
                Y=Y,
                epsilon=self.epsilon,
                beta=beta,
                lambda1=self.lambda1,
                lambda2=self.lambda2,
                weight=self._weight,
                beta_max=self.beta_max,
                max_approx=self.max_approx,
                max_iterations=self.max_iterations,
                debug=self.debug,
            )
        finally:
            set_threads(threads)
        alpha = np.concatenate(
            (y - np.sum(alpha * x, dtype=x.dtype, keepdims=True), alpha)
        )
        self._alpha = alpha
        if self._normalise:
            y = self._y
            if self._logit:
                y = limited_logit(y)
            alpha2 = self._scale.unscale_model(alpha)
            alpha2[0] = y - np.sum(self._x * alpha2[1:])
            self._coefficients = alpha2
        else:
            self._coefficients = alpha
        return self

    def get_params(self, normalised: bool = False) -> np.ndarray:
        """Get the explanation as the coefficients of a linear model (approximating the black box model).

        Args:
            normalised (bool, optional): If the data is normalised within SLISE, return a linear model fitting the normalised data. Defaults to False.

        Returns:
            np.ndarray: The coefficients of the linear model (the first scalar in the vector is the intercept).
        """
        warn("Use `coefficients` instead of `get_params().", SliseWarning)
        return self._alpha if normalised else self._coefficients

    @property
    def coefficients(self) -> np.ndarray:
        """Get the explanation as the coefficients of a linear model (approximating the black box model).

        Returns:
            np.ndarray: The coefficients of the linear model (the first scalar in the vector is the intercept).
        """
        if self._coefficients is None:
            warn("Fit an explanation before retrieving coefficients", SliseWarning)
        return self._coefficients

    def normalised(self, all_columns: bool = True) -> Optional[np.ndarray]:
        """Get coefficients for normalised data (if the data is normalised within SLISE).

        Args:
            all_columns (bool, optional): Add coefficients for constant columns. Defaults to True.

        Returns:
            Optional[np.ndarray]: The normalised coefficients or None.
        """
        if self._alpha is None:
            warn("Fit an explanation before retrieving coefficients", SliseWarning)
        if self._normalise:
            if all_columns:
                return add_constant_columns(self._alpha, self._scale.columns, True)
            else:
                return self._alpha
        else:
            return None

    @property
    def scaled_epsilon(self) -> float:
        """Espilon fitting unnormalised data (if the data is normalised).

        Returns:
            float: Scaled epsilon.
        """
        if self._normalise:
            return self.epsilon * self._scale.y_scale
        else:
            return self.epsilon

    def predict(self, X: Union[np.ndarray, None] = None) -> np.ndarray:
        """Use the approximating linear model to predict new outcomes.

        Args:
            X (Union[np.ndarray, None], optional): Sata matrix to predict, or None for using the fitted dataset. Defaults to None.

        Returns:
            np.ndarray: Prediction vector.
        """
        if X is None:
            Y = mat_mul_inter(self._X, self.coefficients)
        else:
            Y = mat_mul_inter(X, self.coefficients)
        if self._logit:
            Y = sigmoid(Y)
        return Y

    def score(
        self, X: Union[np.ndarray, None] = None, Y: Union[np.ndarray, None] = None
    ) -> float:
        """Calculate the loss. Lower is better and it should usually be negative (unless the regularisation is very (/too?) strong).

        Args:
            X (Union[np.ndarray, None], optional): Data matrix, or None for using the fitted dataset. Defaults to None.
            Y (Union[np.ndarray, None], optional): Response vector, or None for using the fitted dataset. Defaults to None.

        Returns:
            float: The loss.
        """
        if self._alpha is None:
            warn("Fit an explanation before calculating the score", SliseWarning)
        x = self._x
        y = self._y
        if self._logit:
            y = limited_logit(y)
        if self._normalise:
            x = self._scale.scale_x(x)
            y = self._scale.scale_y(y)
        if X is None or Y is None:
            X = self._X2
            Y = self._Y2
        else:
            if self._logit:
                Y = limited_logit(Y)
            if self._normalise:
                X = self._scale.scale_x(X)
                Y = self._scale.scale_y(Y)
        X = X - x[None, :]
        Y = Y - y
        return loss_sharp(
            self._alpha[1:],
            X,
            Y,
            self.epsilon,
            self.lambda1,
            self.lambda2,
            self._weight,
        )

    loss = score
    value = score

    def subset(
        self, X: Union[np.ndarray, None] = None, Y: Union[np.ndarray, None] = None
    ) -> np.ndarray:
        """Get the subset / neighbourhood used for the approximation (explanation).

        Args:
            X (Union[np.ndarray, None], optional): Data matrix, or None for using the fitted dataset. Defaults to None.
            Y (Union[np.ndarray, None], optional): Response vector, or None for using the fitted dataset. Defaults to None.

        Returns:
            np.ndarray: The subset as a boolean mask.
        """
        if X is None or Y is None:
            X = self._X
            Y = self._Y
        if self._logit:
            Y = limited_logit(Y)
        res = mat_mul_inter(X, self.coefficients) - Y
        return res**2 < self.scaled_epsilon**2

    def get_terms(
        self, normalised: bool = False, x: Union[None, np.ndarray] = None
    ) -> np.ndarray:
        """Get the "terms" of different variables on the outcome.
            The terms are the (normalised) coefficients times the (normalised) values.

        Args:
            normalised (bool, optional): Return the normalised terms (if normalisation is used). Defaults to False.
            x (Union[None, np.ndarray], optional): The item to calculate the terms for (uses the explained item if None). Defaults to None.

        Returns:
            np.ndarray: The terms vector.
        """
        if x is None:
            x = self._x
        if normalised and self._normalise:
            x = add_constant_columns(self._scale.scale_x(x), self._scale.columns, False)
            return add_intercept_column(x) * self.coefficients
        else:
            return add_intercept_column(x) * self.coefficients

    get_impact = get_terms

    def print(
        self,
        variables: Union[List[str], None] = None,
        classes: Union[List[str], None] = None,
        num_var: int = 10,
        decimals: int = 3,
    ):
        """Print the current explanation.

        Args:
            variables (Union[List[str], None], optional): Names of the (columns/) variables. Defaults to None.
            classes (Union[List[str], None], optional): Names of the classes, if explaining a classifier. Defaults to None.
            num_var (int, optional): Exclude zero weights if there are too many variables. Defaults to 10.
            decimals (int, optional): Precision to use for printing. Defaults to 3.
        """
        print_slise(
            self.coefficients,
            True,
            self.subset(),
            self.score(),
            self.scaled_epsilon,
            variables,
            "SLISE Explanation",
            decimals,
            num_var,
            unscaled=self._x,
            unscaled_y=self._y,
            terms=self.get_terms(False),
            scaled=None if self._scale is None else self._scale.scale_x(self._x, False),
            alpha=self.normalised(),
            scaled_terms=None if self._scale is None else self.get_terms(True),
            classes=classes,
            unscaled_preds=self._Y,
            logit=self._logit,
        )

    def plot_2d(
        self,
        title: str = "SLISE Explanation",
        label_x: str = "x",
        label_y: str = "y",
        decimals: int = 3,
        fig: Union[Figure, None] = None,
    ) -> SliseRegression:
        """Plot the explanation in a 2D scatter plot (where the explained item is marked) with a line for the approximating model.

        Args:
            title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
            label_x (str, optional): x-axis label. Defaults to "x".
            label_y (str, optional): Y-axis label. Defaults to "y".
            decimals (int, optional): Number of decimals when writing numbers. Defaults to 3.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.

        Raises:
            SliseException: If the data has too many dimensions.
        """
        plot_2d(
            self._X,
            self._Y,
            self.coefficients,
            self.scaled_epsilon,
            self._x,
            self._y,
            self._logit,
            title,
            label_x,
            label_y,
            decimals,
            fig,
        )

    def plot_image(
        self,
        width: int,
        height: int,
        saturated: bool = True,
        title: str = "SLISE Explanation",
        classes: Union[List, str, None] = None,
        decimals: int = 3,
        fig: Union[Figure, None] = None,
    ) -> SliseExplainer:
        """Plot the current explanation for a black and white image (e.g. MNIST).

        Args:
            width (int): Width of the image.
            height (int): Height of the image.
            saturated (bool, optional): Should the explanation be more saturated. Defaults to True.
            title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
            classes (Union[List, str, None], optional): List of class names (first the negative, then the positive), or a single (positive) class name. Defaults to None.
            decimals (int, optional): Number of decimals to write. Defaults to 3.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
        """
        plot_image(
            self._x,
            self._y,
            self._Y,
            self.coefficients,
            width,
            height,
            saturated,
            title,
            classes,
            decimals,
            fig,
        )

    def plot_dist(
        self,
        title: str = "SLISE Explanation",
        variables: list = None,
        order: Union[None, int, Sequence[int]] = None,
        decimals: int = 3,
        fig: Union[Figure, None] = None,
    ) -> SliseExplainer:
        """Plot the current explanation with density distributions for the dataset and a barplot for the model.

        The barplot contains both the approximating linear model (where the weights can be loosely interpreted as the importance of the different variables and their sign) and the "terms", which is the (scaled) model time the (scaled) item values.
        The terms demonstrates how the explained item interacts with the approximating linear model, since a negative weight times a negative value actually supports a positive prediction.

        Args:
            title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
            variables (list, optional): Names for the variables. Defaults to None.
            order (Union[None, int, Sequence[int]], optional): Select variables (None: all, int: largest, selected). Defaults to all.
            decimals (int, optional): Number of decimals to write. Defaults to 3.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
        """
        plot_dist(
            X=self._X,
            Y=self._Y,
            model=self.coefficients,
            subset=self.subset(),
            alpha=self.normalised(),
            x=self._x,
            y=self._y,
            terms=self.get_terms(False),
            norm_terms=self.get_terms(True) if self._normalise else None,
            title=title,
            variables=variables,
            order=order,
            decimals=decimals,
            fig=fig,
        )

    def plot_subset(
        self,
        title: str = "Prediction Distribution",
        decimals: int = 0,
        fig: Union[Figure, None] = None,
    ):
        """Plot a density distributions for predictions and the predictions of the subset

        Args:
            title (str, optional): Title of the plot. Defaults to "Prediction Distribution".
            decimals (int, optional): Number of decimals when writing the subset size. Defaults to 0.
            fig (Union[Figure, None], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
        """
        plot_dist_single(self._Y, self.subset(), self._y, title, decimals, fig)



import numpy as np
from slise.data import (
    pca_simple,
    pca_invert,
    pca_rotate,
    pca_invert_model,
    pca_rotate_model,
    add_intercept_column,
    remove_intercept_column,
    remove_constant_columns,
    add_constant_columns,
    normalise_robust,
    scale_same,
    unscale_model,
)
from slise.utils import mat_mul_inter

from .utils import data_create2, data_create


def test_scaling():
    print("Testing scaling")
    for i in (4, 6, 8):
        X, Y = data_create(i * 30, i, 100000)
        X2, center, scale = normalise_robust(X)
        assert np.allclose(scale_same(X, center, scale), X2)
        assert np.allclose(X2[0,], scale_same(X[0,], center, scale))
        X3 = add_intercept_column(X2)
        assert np.allclose(X2, remove_intercept_column(X3))
        X4, mask = remove_constant_columns(X3)
        assert np.allclose(X2, X4)
        assert np.allclose(mask, np.array([False] + [True] * i))
        assert np.allclose(X3[:, 1:], add_constant_columns(X2, mask)[:, 1:])
        Y2, center2, scale2 = normalise_robust(Y)
        assert np.allclose(scale_same(Y, center2, scale2), Y2)
        assert np.allclose(scale_same(Y[0], center2, scale2), Y2[0])


def test_model_scaling():
    print("Testing model scaling")
    for i in (4, 6, 8):
        X, Y, model2 = data_create2(i * 30, i)
        X2, x_center, x_scale = normalise_robust(X)
        Y2, y_center, y_scale = normalise_robust(Y)
        model2 = np.random.normal(size=i)
        model = unscale_model(model2, x_center, x_scale, y_center, y_scale)
        Z1 = mat_mul_inter(X, model)
        Z2 = mat_mul_inter(X2, model2)
        Z3 = scale_same(Z1, y_center, y_scale)
        assert np.allclose(Z2, Z3), f"Max Diff {np.max(np.abs(Z2 - Z3))}"
        model2 = np.random.normal(size=i + 1)
        model = unscale_model(model2, x_center, x_scale, y_center, y_scale)
        Z1 = mat_mul_inter(X, model)
        Z2 = mat_mul_inter(X2, model2)
        Z3 = scale_same(Z1, y_center, y_scale)
        assert np.allclose(Z2, Z3), f"Max Diff {np.max(np.abs(Z2 - Z3))}"


def test_pca():
    print("Testing pca")
    X = np.random.normal(size=[20, 5])
    X2, v = pca_simple(X, 5)
    assert np.allclose(X, pca_invert(X2, v))
    X3 = np.concatenate((X * 2, X), 1)
    assert pca_simple(X3, 10)[1].shape == (5, 10)
    assert np.allclose(X[0, :], pca_invert(pca_rotate(X[0, :], v), v))
    mod = np.random.normal(size=5)
    assert np.allclose(X @ mod, X2 @ pca_rotate_model(mod, v))
    assert np.allclose(X @ pca_invert_model(mod, v), X2 @ mod)
    X4, v = pca_simple(X.T, 4)



"""
This script contains the loss functions and optimisation functions for SLISE.
"""

from math import log
from typing import Callable, Optional, Tuple
from warnings import catch_warnings, warn

import numpy as np
from lbfgs import LBFGSError, fmin_lbfgs
from numba import jit, get_num_threads, set_num_threads, threading_layer, float64
from scipy.optimize import brentq

from slise.utils import (
    SliseWarning,
    dlog_sigmoid,
    log_sigmoid,
    log_sum_special,
    mat_mul_inter,
    sigmoid,
)


def loss_smooth(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    beta: float = 100,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
) -> float:
    """Smoothed version of the SLISE loss ([slise.optimisation.loss_sharp][]).

    Args:
        alpha (np.ndarray): Linear model coefficients.
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        epsilon (float): Error tolerance.
        beta (float, optional): Sigmoid steepness. Defaults to 100.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 0.
        lambda2 (float, optional): Ridge/L2 regularisation coefficient. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.

    Returns:
        float: Loss value.
    """
    epsilon *= epsilon
    residual2 = ((X @ alpha) - Y) ** 2
    subset = sigmoid(beta * (epsilon - residual2))
    loss = 0.0
    if weight is None:
        residual2 = np.minimum(0, residual2 - epsilon * len(Y))
        loss += np.sum(subset * residual2) / len(Y)
    else:
        sumw = np.sum(weight)
        residual2 = np.minimum(0, residual2 - epsilon * sumw)
        loss += np.sum(subset * residual2 * weight) / sumw
    if lambda1 > 0:
        loss += lambda1 * np.sum(np.abs(alpha))
    if lambda2 > 0:
        loss += lambda2 * np.sum(alpha * alpha)
    return loss


def loss_residuals(
    alpha: np.ndarray,
    residuals2: np.ndarray,
    epsilon2: float,
    beta: float = 100.0,
    lambda1: float = 0.0,
    lambda2: float = 0.0,
    weight: Optional[np.ndarray] = None,
) -> float:
    """Smoothed version of the SLISE loss ([slise.optimisation.loss_smooth][]), that takes already calculated residuals.

    Args:
        alpha (np.ndarray): Linear model coefficients.
        residuals2 (np.ndarray): Squared residuals.
        epsilon2 (float): Squared error tolerance.
        beta (float, optional): Sigmoid steepness. Defaults to 100.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 0.
        lambda2 (float, optional): Ridge/L2 regularisation coefficient. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.

    Returns:
        float: Loss value.
    """
    alpha = np.ascontiguousarray(alpha, dtype=np.float64)
    residuals2 = np.ascontiguousarray(residuals2, dtype=np.float64)
    lambda1 = float(lambda1)
    lambda2 = float(lambda2)
    epsilon2 = float(epsilon2)
    beta = float(beta)
    if weight is None:
        return _loss_residuals(alpha, residuals2, epsilon2, beta, lambda1, lambda2)
    else:
        weight = np.ascontiguousarray(weight, dtype=np.float64)
        return _loss_residualsw(
            alpha, residuals2, epsilon2, beta, lambda1, lambda2, weight
        )


@jit(
    (float64[::1], float64[::1], float64, float64, float64, float64),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _loss_residuals(
    alpha: np.ndarray,
    residuals2: np.ndarray,
    epsilon2: float,
    beta: float = 100,
    lambda1: float = 0,
    lambda2: float = 0,
) -> float:
    """
    Smoothed version of the SLISE loss, that takes already calculated squared residuals.
    _This function is sped up with numba._
    """
    subset = 1 / (1 + np.exp(-beta * (epsilon2 - residuals2)))  # Sigmoid
    residual2 = np.minimum(0, residuals2 - epsilon2 * len(residuals2))
    loss = np.sum(subset * residual2) / len(residuals2)
    if lambda1 > 0:
        loss = loss + lambda1 * np.sum(np.abs(alpha))
    if lambda2 > 0:
        loss = loss + lambda2 * np.sum(alpha * alpha)
    return loss


@jit(
    (
        float64[::1],
        float64[::1],
        float64,
        float64,
        float64,
        float64,
        float64[::1],
    ),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _loss_residualsw(
    alpha: np.ndarray,
    residuals2: np.ndarray,
    epsilon2: float,
    beta: float,
    lambda1: float,
    lambda2: float,
    weight: np.ndarray,
) -> float:
    """
    Smoothed version of the SLISE loss, that takes already calculated squared residuals.
    _This function is sped up with numba._
    """
    subset = 1 / (1 + np.exp(-beta * (epsilon2 - residuals2)))  # Sigmoid
    sumw = np.sum(weight)
    residual2 = np.minimum(0, residuals2 - epsilon2 * sumw)
    loss = np.sum(subset * residual2 * weight) / sumw
    if lambda1 > 0:
        loss = loss + lambda1 * np.sum(np.abs(alpha))
    if lambda2 > 0:
        loss = loss + lambda2 * np.sum(alpha * alpha)
    return loss


def loss_sharp(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
) -> float:
    """Exact version (no sigmoid smoothing) of the SLISE loss.

    Args:
        alpha (np.ndarray): Linear model coefficients.
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        epsilon (float): Error tolerance.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 0.
        lambda2 (float, optional): Ridge/L2 regularisation coefficient. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.

    Returns:
        float: Loss value.
    """
    epsilon *= epsilon
    residual2 = (Y - mat_mul_inter(X, alpha)) ** 2
    if weight is None:
        loss = np.sum(residual2[residual2 <= epsilon] - (epsilon * len(Y))) / len(Y)
    else:
        sumw = np.sum(weight)
        mask = residual2 <= epsilon
        loss = np.sum((residual2[mask] - (epsilon * sumw)) * weight[mask]) / sumw
    if lambda1 > 0:
        loss += lambda1 * np.sum(np.abs(alpha))
    if lambda2 > 0:
        loss += lambda2 * np.sum(alpha * alpha)
    return loss


def loss_grad(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    beta: float,
    lambda1: float = 0.0,
    lambda2: float = 0.0,
    weight: Optional[np.ndarray] = None,
) -> Tuple[float, np.ndarray]:
    """Smoothed version of the SLISE loss ([slise.optimisation.loss_smooth][]), that also calculates the gradient.

    Args:
        alpha (np.ndarray): Linear model coefficients.
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        epsilon (float): Error tolerance.
        beta (float): Sigmoid steepness.
        lambda1 (float): Lasso/L1 regularisation coefficient. Defaults to 0.0.
        lambda2 (float): Ridge/L2 regularisation coefficient. Defaults to 0.0.
        weight (Optional[np.ndarray]): Weight vector for the data items. Defaults to None.

    Returns:
        Tuple[float, np.ndarray]: Loss value and gradient vector.
    """
    alpha = np.ascontiguousarray(alpha, dtype=np.float64)
    X = np.ascontiguousarray(X, dtype=np.float64)
    Y = np.ascontiguousarray(Y, dtype=np.float64)
    assert X.shape[0] == len(Y), f"Different lengths {X.shape[0]} != {len(Y)}"
    assert X.shape[1] == len(alpha), f"Different lengths {X.shape[0]} != {len(alpha)}"
    lambda1 = float(lambda1)
    lambda2 = float(lambda2)
    epsilon = float(epsilon)
    beta = float(beta)
    if weight is None:
        loss, grad = _loss_grad(alpha, X, Y, epsilon, beta, lambda2)
    else:
        weight = np.ascontiguousarray(weight, dtype=np.float64)
        assert Y.shape == weight.shape, f"Different shapes {Y.shape} != {weight.shape}"
        loss, grad = _loss_gradw(alpha, X, Y, epsilon, beta, lambda2, weight)
    if lambda1 > 0:
        loss = loss + lambda1 * np.sum(np.abs(alpha))
        grad = grad + lambda1 * np.sign(alpha)
    return loss, grad


def loss_numba(*args, **kwargs):
    warn("Use `loss_grad` instead of `loss_numba`", DeprecationWarning)
    return loss_grad(*args, **kwargs)


@jit(
    (float64[::1], float64[:, ::1], float64[::1], float64, float64, float64),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _loss_grad(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    beta: float,
    lambda2: float,
) -> Tuple[float, np.ndarray]:
    """Smoothed version of the SLISE loss that also calculates the gradient.
    _This function is sped up with numba._
    """
    epsilon *= epsilon
    residuals = (X @ alpha) - Y
    residual2 = residuals**2
    n = residuals.dtype.type(len(Y))
    # Loss
    subset = 1 / (1 + np.exp(-beta * (epsilon - residual2)))  # Sigmoid
    residual2 = np.minimum(0, residual2 - (epsilon * n))
    loss = np.sum(subset * residual2) / n
    # Gradient
    grad = (
        np.expand_dims(
            subset
            * residuals
            * (2.0 / n - residual2 * (2.0 * beta / n) * (1.0 - subset))
            * (residual2 < 0.0).astype(X.dtype),
            0,
        )
        @ X
    )
    # Lambda
    if lambda2 > 0:
        loss = loss + lambda2 * np.sum(alpha * alpha)
        grad = grad + (lambda2 * 2) * alpha
    return loss, grad


@jit(
    (
        float64[::1],
        float64[:, ::1],
        float64[::1],
        float64,
        float64,
        float64,
        float64[::1],
    ),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _loss_gradw(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    beta: float,
    lambda2: float,
    weight: np.ndarray,
) -> Tuple[float, np.ndarray]:
    """Smoothed version of the SLISE loss, that also calculates the gradient.
    _This function is sped up with numba._
    """
    epsilon *= epsilon
    residuals = (X @ alpha) - Y
    residual2 = residuals**2
    n = np.sum(weight)
    # Loss
    subset = 1 / (1 + np.exp(-beta * (epsilon - residual2)))  # Sigmoid
    residual2 = np.minimum(0, residual2 - (epsilon * n))
    loss = np.sum(subset * residual2 * weight) / n
    # Gradient
    grad = (
        np.expand_dims(
            subset
            * residuals
            * weight
            * (2.0 / n - residual2 * (2.0 * beta / n) * (1.0 - subset))
            * (residual2 < 0.0).astype(X.dtype),
            0,
        )
        @ X
    )
    # Lambda
    if lambda2 > 0:
        loss = loss + lambda2 * np.sum(alpha * alpha)
        grad = grad + (lambda2 * 2) * alpha
    return loss, grad


@jit(
    (float64[::1], float64[:, ::1], float64[::1], float64),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _ridge_numba(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    lambda2: float = 0.0,
) -> Tuple[float, np.ndarray]:
    """
    Ridge regression (OLS + L2) loss, that also calculates the gradient.
    _This function is sped up with numba._
    """
    distances = (X @ alpha) - Y
    loss = np.sum(distances**2) / 2
    grad = np.expand_dims(distances, 0) @ X
    if lambda2 > 0.0:
        loss = loss + lambda2 * np.sum(alpha**2) / 2
        grad = grad + lambda2 * alpha
    return loss, grad


@jit(
    (float64[::1], float64[:, ::1], float64[::1], float64, float64[::1]),
    nopython=True,
    fastmath=True,
    parallel=True,
    cache=True,
    nogil=True,
    boundscheck=False,
)
def _ridge_numbaw(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    lambda2: float,
    weight: np.ndarray,
) -> Tuple[float, np.ndarray]:
    """
    Ridge regression (OLS + L2) loss, that also calculates the gradient.
    _This function is sped up with numba._
    """
    distances = ((X @ alpha) - Y) * weight
    loss = np.sum(distances**2) / 2
    grad = np.expand_dims(distances, 0) @ X
    if lambda2 > 0.0:
        loss = loss + lambda2 * np.sum(alpha**2) / 2
        grad = grad + lambda2 * alpha
    return loss, grad


def owlqn(
    loss_grad_fn: Callable[[np.ndarray], Tuple[float, np.ndarray]],
    x0: np.ndarray,
    lambda1: float = 0,
    max_iterations: int = 200,
    **kwargs,
) -> np.ndarray:
    """Wrapper around owlqn that converts max_iter errors to warnings (see `lbfgs.fmin_lbfgs` from PyLBFGS).

    Args:
        loss_grad_fn (Callable[[np.ndarray], Tuple[float, np.ndarray]]): Function that calculates the loss and gradient.
        x0 (np.ndarray): Initial vector to be optimised.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 1e-6.
        max_iterations (int, optional): Maximum number of optimisation steps. Defaults to 200.

    Returns:
        np.ndarray: Optimised vector.
    """

    def f(x: np.ndarray, gradient: np.ndarray) -> float:
        loss, grad = loss_grad_fn(x)
        gradient[:] = grad
        return loss

    try:  # PyLBFGS throws an error if max_iterations is exceeded, this is a workaround to convert it into a warning

        def p(x, g, fx, xnorm, gnorm, step, k, num_eval, *args):
            if k >= max_iterations:
                x0[:] = x

        x0 = fmin_lbfgs(
            f=f,
            x0=x0,
            progress=p,
            orthantwise_c=lambda1,
            max_iterations=max_iterations,
            line_search="wolfe" if lambda1 > 0 else "default",
            **kwargs,
        )
    except LBFGSError as error:
        if (
            error.args[0]
            != "The algorithm routine reaches the maximum number of iterations."
        ):
            raise error
        else:
            warn(
                "LBFGS optimisation reaches the maximum number of iterations.",
                SliseWarning,
            )
    return x0


def regularised_regression(
    X: np.ndarray,
    Y: np.ndarray,
    lambda1: float = 1e-6,
    lambda2: float = 1e-6,
    weight: Optional[np.ndarray] = None,
    max_iterations: int = 200,
) -> np.ndarray:
    """Train a linear regression model with lasso (L1) and/or ridge (L2) regularisation.

    Args:
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 1e-6.
        lambda2 (float, optional): Ridge/L2 regularisation coefficient. Defaults to 1e-6.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
        max_iterations (int, optional): Maximum number of optimisation steps. Defaults to 200.

    Returns:
        np.ndarray: The coefficients of the linear model.
    """
    X = np.ascontiguousarray(X, dtype=np.float64)
    Y = np.ascontiguousarray(Y, dtype=np.float64)
    lambda1 = float(lambda1)
    lambda2 = float(lambda2)
    assert X.shape[0] == len(Y), f"Different lengths {X.shape[0]} != {len(Y)}"
    if weight is None:
        lf = lambda alpha: _ridge_numba(alpha, X, Y, lambda2)  # noqa: E731
    else:
        weight = np.ascontiguousarray(weight, dtype=np.float64)
        assert Y.shape == weight.shape, f"Different shapes {Y.shape} != {weight.shape}"
        lf = lambda alpha: _ridge_numbaw(alpha, X, Y, lambda2, weight)  # noqa: E731
    return owlqn(lf, np.zeros(X.shape[1], dtype=np.float64), lambda1, max_iterations)


def optimise_loss(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float = 0.1,
    beta: float = 100,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
    max_iterations: int = 200,
) -> np.ndarray:
    """Optimise a smoothed SLISE loss with `owl-qn`.

    Args:
        alpha (np.ndarray): Linear model coefficients.
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Target vector
        epsilon (float, optional): Error tolerance. Defaults to 0.1.
        beta (float, optional): Sigmoid steepness. Defaults to 100.
        lambda1 (float, optional): LASSO/L1 regularisation coefficient. Defaults to 1e-6.
        lambda2 (float, optional): Ridge/L2 regularisation coefficient. Defaults to 1e-6.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
        max_iterations (int, optional): Maximum number of optimisation steps. Defaults to 200.

    Returns:
        np.ndarray: The coefficients of the linear model.
    """
    alpha = np.ascontiguousarray(alpha, dtype=np.float64)
    X = np.ascontiguousarray(X, dtype=np.float64)
    Y = np.ascontiguousarray(Y, dtype=np.float64)
    assert X.shape[0] == len(Y), f"Different lengths {X.shape[0]} != {len(Y)}"
    assert X.shape[1] == len(alpha), f"Different lengths {X.shape[0]} != {len(alpha)}"
    lambda1 = float(lambda1)
    lambda2 = float(lambda2)
    epsilon = float(epsilon)
    beta = float(beta)
    if weight is None:
        lf = lambda alpha: _loss_grad(alpha, X, Y, epsilon, beta, lambda2)  # noqa: E731
    else:
        weight = np.ascontiguousarray(weight, dtype=np.float64)
        assert Y.shape == weight.shape, f"Different shapes {Y.shape} != {weight.shape}"
        lf = lambda alpha: _loss_gradw(alpha, X, Y, epsilon, beta, lambda2, weight)  # noqa: E731
    return owlqn(lf, alpha, lambda1, max_iterations)


def log_approximation_ratio(
    residuals2: np.ndarray,
    epsilon2: float,
    beta1: float,
    beta2: float,
    weight: Optional[np.ndarray] = None,
) -> float:
    """Calculate log(K), where K is the approximation ratio between two smoothed losses.

    Args:
        residuals2 (np.ndarray): Squared residuals.
        epsilon2 (float): Squared error tolerance.
        beta1 (float): Old sigmoid steepness.
        beta2 (float): New sigmoid steepness.
        weight (Optional[np.ndarray], optional): Weight vector. Defaults to None.

    Returns:
        float: log of the approximation ratio between `beta1` and `beta2` for the current solution.
    """
    if beta1 >= beta2:
        return 0
    log_f = lambda r, beta: log_sigmoid(beta * (epsilon2 - r))  # noqa: E731
    dlog_g = lambda r: -beta1 * dlog_sigmoid(  # noqa: E731
        beta1 * (epsilon2 - r)
    ) + beta2 * dlog_sigmoid(beta2 * (epsilon2 - r))
    if dlog_g(0) < 0:
        a = brentq(dlog_g, 0, epsilon2)
        log_k = min(
            log_f(0, beta1) - log_f(0, beta2), log_f(a, beta1) - log_f(a, beta2)
        )
    else:
        log_k = log_f(0, beta1) - log_f(0, beta2)
    if weight is None:
        phi = np.maximum(0, epsilon2 - residuals2 / len(residuals2))
    else:
        phi = np.maximum(0, epsilon2 - residuals2 / np.sum(weight)) * weight
    log_K = (
        log_sum_special(log_f(residuals2, beta1), phi)
        - log_k
        - log_sum_special(log_f(residuals2, beta2), phi)
    )
    return log_K


def next_beta(
    residuals2: np.ndarray,
    epsilon2: float = 0.01,
    beta: float = 0.0,
    weight: Optional[np.ndarray] = None,
    beta_max: float = 2500,
    log_max_approx: float = 0.14,
    min_beta_step: float = 0.0005,
) -> float:
    """Calculate the next beta for the graduated optimisation.

    Args:
        residuals2 (np.ndarray): Squared residuals.
        epsilon2 (float): Squared error tolerance. Defaults to 0.01.
        beta (float): Sigmoid steepness. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector. Defaults to None.
        beta_max (float, optional): Maximum `beta`. Defaults to 2500.
        log_max_approx (float, optional): Log-maximum approximation ratio. Defaults to 0.14.
        min_beta_step (float, optional): Minimum increase of `beta`. Defaults to 0.0005.

    Returns:
        float: next `beta`.
    """
    if beta >= beta_max:
        return beta
    log_approx = log_approximation_ratio(residuals2, epsilon2, beta, beta_max, weight)
    if log_approx <= log_max_approx:
        return beta_max
    else:
        f = (  # noqa: E731
            lambda b: log_approximation_ratio(residuals2, epsilon2, beta, b, weight)
            - log_max_approx
        )
        beta_min = beta + min_beta_step * (beta_max + beta)
        return max(brentq(f, beta, beta_max), beta_min)


def matching_epsilon(
    residuals2: np.ndarray,
    epsilon2: float,
    beta: float,
    weight: Optional[np.ndarray] = None,
) -> float:
    """Approximately calculate the epsilon that minimises the approximation ratio to the exact loss.

    Args:
        residuals2 (np.ndarray): Squared residuals.
        epsilon2 (float): Squared error tolerance.
        beta (float): Sigmoid steepness.
        weight (Optional[np.ndarray], optional): Weight vector. Defaults to None.

    Returns:
        float: (Approximatively) optimal epsilon for the exact loss (for the current solution).
    """
    if weight is None:
        residuals2 = np.sort(residuals2)
        loss = sigmoid(beta * (epsilon2 - residuals2))
        i = np.argmax(np.arange(1, 1 + len(residuals2)) * loss)
        return residuals2[i] ** 0.5
    else:
        order = np.argsort(residuals2)
        residuals2 = residuals2[order]
        loss = sigmoid(beta * (epsilon2 - residuals2))
        i = np.argmax(np.cumsum(weight[order]) * loss)
        return residuals2[i] ** 0.5


def _debug_log(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float = 0.1,
    beta: float = 0,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
):
    """
    Print the log statement for a graduated optimisation step.
    """
    residuals = (X @ alpha - Y) ** 2
    loss = loss_sharp(alpha, X, Y, epsilon, lambda1, lambda2, weight)
    bloss = loss_residuals(alpha, residuals, epsilon**2, beta, lambda1, lambda2, weight)
    epss = matching_epsilon(residuals, epsilon**2, beta, weight)
    beta = beta * epsilon**2
    print(
        f"beta: {beta:5.3f}    epsilon*: {epss:.3f}    Loss: {loss:6.2f}    B-Loss: {bloss:6.2f}"
    )


@np.errstate(over="ignore")
def graduated_optimisation(
    alpha: np.ndarray,
    X: np.ndarray,
    Y: np.ndarray,
    epsilon: float,
    beta: float = 0,
    lambda1: float = 0,
    lambda2: float = 0,
    weight: Optional[np.ndarray] = None,
    beta_max: float = 20,
    max_approx: float = 1.15,
    max_iterations: int = 200,
    debug: bool = False,
) -> np.ndarray:
    """Optimise `alpha` using graduated optimisation.

    Args:
        alpha (np.ndarray): Initial linear model coefficients.
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        epsilon (float): Error tolerance.
        beta (float, optional): Initial sigmoid steepness. Defaults to 0.
        lambda1 (float, optional): L1 regularisation strength. Defaults to 0.
        lambda2 (float, optional): L2 regularisation strength. Defaults to 0.
        weight (Optional[np.ndarray], optional): Weight vector for the data items. Defaults to None.
        beta_max (float, optional): Maximum sigmoid steepness (the final beta). Defaults to 20.
        max_approx (float, optional): Target approximation ratio when increasing beta. Defaults to 1.15.
        max_iterations (int, optional): Maximum number of iterations for owl-qn. Defaults to 200.
        debug (bool, optional): Print debug logs after each optimisation step. Defaults to False.

    Returns:
        np.ndarray: Optimised `alpha`.
    """
    X = np.ascontiguousarray(X, dtype=np.float64)
    Y = np.ascontiguousarray(Y, dtype=np.float64)
    if weight is not None:
        weight = np.ascontiguousarray(weight, dtype=np.float64)
    beta_max = beta_max / epsilon**2
    max_approx = log(max_approx)
    with catch_warnings(record=True) as w:
        while beta < beta_max:
            alpha = optimise_loss(
                alpha, X, Y, epsilon, beta, lambda1, lambda2, weight, max_iterations
            )
            if debug:
                _debug_log(alpha, X, Y, epsilon, beta, lambda1, lambda2, weight)
            beta = next_beta(
                (X @ alpha - Y) ** 2, epsilon**2, beta, weight, beta_max, max_approx
            )
    alpha = optimise_loss(
        alpha, X, Y, epsilon, beta, lambda1, lambda2, weight, max_iterations * 4
    )
    if debug:
        _debug_log(alpha, X, Y, epsilon, beta, lambda1, lambda2, weight)
        if w:
            print("Warnings from intermediate steps:", w)
    return alpha


def set_threads(num: int = -1) -> int:
    """Set the number of numba threads.

    Args:
        num (int, optional): The number of threads (or -1 to keep the old value). Defaults to -1.

    Returns:
        int: The old number of theads (or -1 if unchanged).
    """
    if num > 0:
        old = get_num_threads()
        if old != num:
            set_num_threads(num)
        return old
    return -1


@jit((float64[::1],), nopython=True, fastmath=True, parallel=True, cache=True)
def _dummy_numba(
    x: np.ndarray,
) -> np.ndarray:
    """
    A dummy function to check the numba compilation (see check_threading_layer).
    """
    return x * x


def check_threading_layer():
    """
    Check which numba threading_layer is active, and warn if it is "workqueue".
    """
    _dummy_numba(np.ones(1))
    try:
        if threading_layer() == "workqueue":
            warn(
                'Using `numba.threading_layer()=="workqueue"` can be devastatingly slow!'
                " See https://numba.pydata.org/numba-doc/latest/user/threading-layer.html for alternatives.",
                SliseWarning,
            )
    except ValueError as e:
        warn(f"Numba: {e}", SliseWarning)



"""
This script contains some utility functions.
"""

from typing import Union

import numpy as np
from scipy.special import expit as sigmoid


class SliseWarning(RuntimeWarning):
    """
    Custom tag for warnings.
    """


class SliseException(Exception):
    """
    Custom tag for exceptions.
    """


def limited_logit(
    p: Union[np.ndarray, float], stab: float = 0.001
) -> Union[np.ndarray, float]:
    """Computes logits from probabilities.

    Args:
        p (Union[np.ndarray, float]): Probability vector or scalar.
        stab (float, optional): Limit p to [stab, 1-stab] for numerical stability. Defaults to 0.001.

    Returns:
        Union[np.ndarray, float]: `logit(clamp(p, stab, 1-stab))`.
    """
    p = np.minimum(1.0 - stab, np.maximum(stab, p))
    return np.log(p / (1.0 - p))


def dsigmoid(x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Derivative of the sigmoid function.

    Args:
        x (Union[np.ndarray, float]): Real vector or scalar.

    Returns:
        Union[np.ndarray, float]: Derivative of `sigmoid(x)`.
    """
    s = sigmoid(x)
    return s * (1 - s)


def log_sigmoid(x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Computes `log(sigmoid(x))` in a numerically stable way.

    Args:
        x (Union[np.ndarray, float]): Real vector or scalar.

    Returns:
        Union[np.ndarray, float]: `log(sigmoid(x))`
    """
    y = -np.sign(x)
    return (y * 0.5 + 0.5) * x - np.log1p(np.exp(y * x))


def dlog_sigmoid(x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Derivative of `log(sigmoid(x))`.

    Args:
        x (Union[np.ndarray, float]): Real vector or scalar.

    Returns:
        Union[np.ndarray, float]: Derivative of `log(sigmoid(x))`
    """
    return 1 - sigmoid(x)


def sparsity(x: Union[np.ndarray, float], treshold: float = 0) -> int:
    """Count the number of `abs(x) > treshold`.

    Args:
        x (Union[np.ndarray, float]): Real vector or scalar.
        treshold (float, optional): Threshold non-inclusive. Defaults to 0.

    Returns:
        int: The number of `abs(x) > treshold`.
    """
    if treshold > 0:
        return np.count_nonzero(np.abs(x) > treshold)
    else:
        return np.count_nonzero(x)


def log_sum_exp(x: np.ndarray) -> float:
    """Computes `log(sum(exp(x)))` in a numerically stable way.

    Args:
        x (np.ndarray): Real vector.

    Returns:
        float: `log(sum(exp(x)))`
    """
    xmax = np.max(x)
    return xmax + np.log(np.sum(np.exp(x - xmax)))


def log_sum_special(x: np.ndarray, y: np.ndarray) -> float:
    """Computes `log(sum(exp(x) * y))` (or `log(sum(exp(x)))` if `all(y == 0)`), in a numerically stable way.

    Args:
        x (np.ndarray): Real vector.
        y (np.ndarray): Real vector.

    Returns:
        float: `log(sum(exp(x) * y))`.
    """
    xmax = np.max(x)
    xexp = np.exp(x - xmax)
    xsum = np.sum(xexp * y)
    if xsum == 0:
        xsum = np.sum(xexp)
    return xmax + np.log(xsum)


def mat_mul_inter(X: np.ndarray, alpha: np.ndarray) -> np.ndarray:
    """Matrix multiplication, but check and handle potential intercepts in `alpha`.

    Args:
        X (np.ndarray): Real matrix or vector.
        alpha (np.ndarray): Real vector.

    Returns:
        np.ndarray: `X @ alpha` or `X @ alpha[1:] + alpha[0]`.
    """
    alpha = np.atleast_1d(alpha)
    if len(X.shape) == 1:
        if len(alpha) == X.size:
            return np.sum(alpha[1:] * X)
        if len(alpha) == X.size + 1:
            return alpha[0] + np.sum(alpha[1:] * X)
        else:
            X = np.reshape(X, X.shape + (1,))
    if len(alpha) == X.shape[1] + 1:
        return X @ alpha[1:] + alpha[0]
    else:
        return X @ alpha



import numpy as np
from pytest import approx
from slise.optimisation import (
    check_threading_layer,
    graduated_optimisation,
    log_approximation_ratio,
    loss_grad,
    loss_sharp,
    loss_smooth,
    matching_epsilon,
    next_beta,
    optimise_loss,
    regularised_regression,
)
from slise.utils import log_sigmoid, log_sum_exp, log_sum_special, sigmoid, sparsity

from .utils import data_create, data_create2, numeric_grad


def test_utils():
    print("Testing util functions")
    x = np.arange(-6, 6)
    assert np.allclose(np.log(sigmoid(x)), log_sigmoid(x))
    assert sparsity(x) == len(x) - 1
    assert sparsity(x, 1.2) == len(x) - 3
    assert log_sum_exp(x) == np.log(np.sum(np.exp(x)))
    assert log_sum_exp(x) == log_sum_special(x, 0)
    assert log_sum_exp(x) == log_sum_special(x, 1)
    assert log_sum_exp(x) + np.log(2) == log_sum_special(x, 2)


def test_loss():
    print("Testing loss functions")
    X, Y = data_create(20, 5)
    w = np.random.uniform(size=20)
    alpha = np.random.normal(size=5)
    assert loss_smooth(alpha, X, Y, 0.1) <= 0
    assert loss_sharp(alpha, X, Y, 0.1) <= 0
    assert loss_grad(alpha, X, Y, 0.1, lambda2=0, beta=0)[0] <= 0
    assert loss_smooth(alpha, X, Y, 10) < 0
    assert loss_sharp(alpha, X, Y, 10) < 0
    assert loss_grad(alpha, X, Y, 10, lambda2=0, beta=0)[0] < 0
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, beta=1000000), loss_sharp(alpha, X, Y, 0.1)
    )
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, lambda1=0.5, beta=1000000),
        loss_sharp(alpha, X, Y, 0.1, lambda1=0.5),
    )
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, lambda2=0.5, beta=1000000),
        loss_sharp(alpha, X, Y, 0.1, lambda2=0.5),
    )
    assert loss_smooth(alpha, X, Y, 0.1, beta=20, lambda1=0.0, lambda2=0.0) == approx(
        loss_grad(alpha, X, Y, 0.1, lambda2=0.0, beta=20)[0], 1e-8
    )
    assert loss_smooth(alpha, X, Y, 0.1, beta=20, lambda1=0.0, lambda2=0.5) == approx(
        loss_grad(alpha, X, Y, 0.1, lambda2=0.5, beta=20)[0], 1e-8
    )
    # With weight
    assert loss_smooth(alpha, X, Y, 0.1, weight=w) <= 0
    assert loss_sharp(alpha, X, Y, 0.1, weight=w) <= 0
    assert loss_grad(alpha, X, Y, 0.1, lambda2=0, beta=0, weight=w)[0] <= 0
    assert loss_smooth(alpha, X, Y, 10, weight=w) < 0
    assert loss_sharp(alpha, X, Y, 10, weight=w) < 0
    assert loss_grad(alpha, X, Y, 10, lambda2=0, beta=0, weight=w)[0] < 0
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, beta=1000000, weight=w),
        loss_sharp(alpha, X, Y, 0.1, weight=w),
    )
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, lambda1=0.5, beta=1000000, weight=w),
        loss_sharp(alpha, X, Y, 0.1, lambda1=0.5, weight=w),
    )
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, lambda2=0.5, beta=1000000, weight=w),
        loss_sharp(alpha, X, Y, 0.1, lambda2=0.5, weight=w),
    )
    assert loss_smooth(alpha, X, Y, 0.1, beta=20, weight=w, lambda2=0.0) == approx(
        loss_grad(alpha, X, Y, 0.1, lambda2=0.0, weight=w, beta=20)[0], 1e-8
    )
    assert loss_smooth(alpha, X, Y, 0.1, beta=20, weight=w, lambda2=0.5) == approx(
        loss_grad(alpha, X, Y, 0.1, lambda2=0.5, weight=w, beta=20)[0], 1e-8
    )


def test_grad_numerically():
    print("Comparing the manual gradient to a numeric gradient")
    for _ in range(3):
        X, Y = data_create(20, 5)
        alpha = np.random.normal(size=5)
        _, grad = loss_grad(alpha, X, Y, 10, lambda2=0, beta=0)
        grad2 = numeric_grad(
            alpha, lambda x: loss_grad(x, X, Y, 10, lambda2=0, beta=0)[0]
        )
        assert np.allclose(grad, grad2, atol=1e-4)
        w = np.random.uniform(size=20)
        _, grad = loss_grad(alpha, X, Y, 10, lambda2=0, beta=0, weight=w)
        grad2 = numeric_grad(
            alpha, lambda x: loss_grad(x, X, Y, 10, lambda2=0, beta=0, weight=w)[0]
        )
        assert np.allclose(grad, grad2, atol=1e-4)


def test_owlqn():
    print("Testing owlqn")
    X, Y = data_create(20, 5)
    alpha = np.random.normal(size=5)
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100
    )
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100, lambda1=0.5)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, lambda1=0.5) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, lambda1=0.5
    )
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100, lambda2=0.5)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, lambda2=0.5) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, lambda2=0.5
    )
    # With weight
    w = np.random.uniform(size=20)
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100, weight=w)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, weight=w) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, weight=w
    )
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100, lambda1=0.5, weight=w)
    assert loss_smooth(
        alpha, X, Y, 0.1, beta=100, lambda1=0.5, weight=w
    ) >= loss_smooth(alpha2, X, Y, 0.1, beta=100, lambda1=0.5, weight=w)
    alpha2 = optimise_loss(alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w)
    assert loss_smooth(
        alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w
    ) >= loss_smooth(alpha2, X, Y, 0.1, beta=100, lambda2=0.5, weight=w)


def test_gradopt():
    print("Testing graduated optimisation")
    X, Y = data_create(20, 5)
    alpha = np.random.normal(size=5)
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, debug=True)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100
    )
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda1=0.5)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, lambda1=0.5) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, lambda1=0.5
    )
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda2=0.5)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, lambda2=0.5) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, lambda2=0.5
    )
    # With weight
    w = np.random.uniform(size=20)
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, weight=w)
    assert loss_smooth(alpha, X, Y, 0.1, beta=100, weight=w) >= loss_smooth(
        alpha2, X, Y, 0.1, beta=100, weight=w
    )
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda1=0.5, weight=w)
    assert loss_smooth(
        alpha, X, Y, 0.1, beta=100, lambda1=0.5, weight=w
    ) >= loss_smooth(alpha2, X, Y, 0.1, beta=100, lambda1=0.5, weight=w)
    alpha2 = graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w)
    assert loss_smooth(
        alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w
    ) >= loss_smooth(alpha2, X, Y, 0.1, beta=100, lambda2=0.5, weight=w)


def test_regres():
    print("Testing regularised regression")
    X, Y, mod = data_create2(20, 5)
    alpha = regularised_regression(X, Y, 1e-10, 1e-10)
    Y2 = X @ alpha
    assert np.allclose(Y, Y2, atol=0.3), f"regreg Y not close: {Y - Y2}"
    assert np.allclose(mod, alpha, atol=0.2), f"regreg alpha not close: {mod - alpha}"
    alpha = regularised_regression(X, Y, 1e-10, 0)
    Y2 = X @ alpha
    assert np.allclose(Y, Y2, atol=0.3), f"Lasso Y not close: {Y - Y2}"
    assert np.allclose(mod, alpha, atol=0.2), f"Lasso alpha not close: {mod - alpha}"
    alpha = regularised_regression(X, Y, 0, 1e-10)
    Y2 = X @ alpha
    assert np.allclose(Y, Y2, atol=0.3), f"Ridge Y not close: {Y - Y2}"
    assert np.allclose(mod, alpha, atol=0.2), f"Ridge alpha not close: {mod - alpha}"
    w = np.ones(20)
    alpha = regularised_regression(X, Y, 1e-10, 1e-10, w)
    Y2 = X @ alpha
    assert np.allclose(Y, Y2, atol=0.3), f"regregw Y not close: {Y - Y2}"
    assert np.allclose(mod, alpha, atol=0.2), f"regregw alpha not close: {mod - alpha}"


def test_weights():
    X, Y, mod = data_create2(20, 5)
    w1 = None
    w2 = np.ones(20)
    w3 = w2 * 2
    X2 = np.concatenate((X, X), 0)
    Y2 = np.concatenate((Y, Y), 0)
    alpha = np.random.normal(size=5)
    assert np.allclose(
        loss_sharp(alpha, X, Y, 0.1, weight=w1),
        loss_sharp(alpha, X, Y, 0.1, weight=w2),
    )
    assert np.allclose(
        loss_sharp(alpha, X2, Y2, 0.1, weight=w1),
        loss_sharp(alpha, X, Y, 0.1, weight=w3),
    )
    assert np.allclose(
        loss_smooth(alpha, X, Y, 0.1, 10, weight=w1),
        loss_smooth(alpha, X, Y, 0.1, 10, weight=w2),
    )
    assert np.allclose(
        loss_smooth(alpha, X2, Y2, 0.1, 10, weight=w1),
        loss_smooth(alpha, X, Y, 0.1, 10, weight=w3),
    )
    assert np.allclose(
        graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w1),
        graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w2),
    )
    assert np.allclose(
        graduated_optimisation(alpha, X2, Y2, 0.1, beta=100, lambda2=0.5, weight=w1),
        graduated_optimisation(alpha, X, Y, 0.1, beta=100, lambda2=0.5, weight=w3),
    )
    assert np.allclose(
        matching_epsilon((X @ alpha - Y) ** 2, 0.01, 10, w1),
        matching_epsilon((X @ alpha - Y) ** 2, 0.01, 10, w2),
    )
    assert np.allclose(
        matching_epsilon((X2 @ alpha - Y2) ** 2, 0.01, 100, w1),
        matching_epsilon((X @ alpha - Y) ** 2, 0.01, 100, w3),
    )
    assert np.allclose(
        next_beta((X @ alpha - Y) ** 2, 0.01, 100, w1),
        next_beta((X @ alpha - Y) ** 2, 0.01, 100, w2),
    )
    assert np.allclose(
        next_beta((X2 @ alpha - Y2) ** 2, 0.01, 100, w1),
        next_beta((X @ alpha - Y) ** 2, 0.01, 100, w3),
    )
    assert np.allclose(
        log_approximation_ratio((X @ alpha - Y) ** 2, 0.01, 1, 100, w1),
        log_approximation_ratio((X @ alpha - Y) ** 2, 0.01, 1, 100, w2),
    )
    assert np.allclose(
        log_approximation_ratio((X2 @ alpha - Y2) ** 2, 0.01, 1, 100, w1),
        log_approximation_ratio((X @ alpha - Y) ** 2, 0.01, 1, 100, w3),
    )
    assert np.allclose(
        regularised_regression(X, Y, 1e-4, 1e-4, weight=w1),
        regularised_regression(X, Y, 1e-4, 1e-4, weight=w2),
    )
    assert np.allclose(
        regularised_regression(X2, Y2, 1e-4, 1e-4, weight=w1),
        regularised_regression(X, Y, 1e-4, 1e-4, weight=w3),
        atol=1e-5,
    )


def test_check_threading_layer():
    check_threading_layer()



from warnings import catch_warnings

import numba
import numpy as np
from pytest import approx
from scipy.special import expit as sigmoid
from slise import explain, regression
from slise.data import add_intercept_column
from slise.initialisation import (
    initialise_candidates,
    initialise_candidates2,
    initialise_fixed,
    initialise_lasso,
    initialise_ols,
    initialise_zeros,
)
from slise.optimisation import loss_smooth
from slise.utils import mat_mul_inter

from .utils import data_create, data_create2


def test_initialise_simple():
    print("Testing old initialisations")
    X, Y = data_create(20, 5)
    w = np.random.uniform(size=20)
    alpha, beta = initialise_lasso(X, Y, 0.1, w)
    alpha, beta = initialise_lasso(X, Y, 0.1)
    assert beta == 0
    assert len(alpha) == 5
    alpha, beta = initialise_zeros(X, Y, 0.1, w)
    alpha, beta = initialise_zeros(X, Y, 0.1)
    assert beta > 0
    assert len(alpha) == 5
    alpha, beta = initialise_ols(X, Y, 0.1, w)
    alpha, beta = initialise_ols(X, Y, 0.1)
    assert beta > 0
    assert len(alpha) == 5
    alpha, beta = initialise_fixed(alpha, X, Y, 0.1, w)
    alpha, beta = initialise_fixed(alpha, X, Y, 0.1)
    assert beta > 0
    assert len(alpha) == 5
    alpha, beta = initialise_fixed((alpha, beta), X, Y, 0.1, w)
    alpha, beta = initialise_fixed((alpha, beta), X, Y, 0.1)
    assert beta > 0
    assert len(alpha) == 5


def test_initialise():
    with catch_warnings(record=True) as w:
        print("Testing initialisation")
        X, Y = data_create(20, 5)
        zero = np.zeros(5)
        alpha, beta = initialise_candidates(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 12)
        zero = np.zeros(12)
        alpha, beta = initialise_candidates(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 11)
        X = add_intercept_column(X)
        zero = np.zeros(12)
        alpha, beta = initialise_candidates(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 8)
        w = np.random.uniform(size=20)
        zero = np.zeros(8)
        alpha, beta = initialise_candidates(X, Y, 0.1, w)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta, weight=w) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta, weight=w
        )


def test_initialise2():
    with catch_warnings(record=True) as w:
        print("Testing initialisation2")
        X, Y = data_create(20, 5)
        zero = np.zeros(5)
        alpha, beta = initialise_candidates2(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 12)
        zero = np.zeros(12)
        alpha, beta = initialise_candidates2(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 11)
        X = add_intercept_column(X)
        zero = np.zeros(12)
        alpha, beta = initialise_candidates2(X, Y, 0.1)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta
        )
        X, Y = data_create(20, 8)
        w = np.random.uniform(size=20)
        zero = np.zeros(8)
        alpha, beta = initialise_candidates2(X, Y, 0.1, w)
        assert beta > 0
        assert loss_smooth(alpha, X, Y, 0.1, beta=beta, weight=w) <= loss_smooth(
            zero, X, Y, 0.1, beta=beta, weight=w
        )


def test_slise_reg():
    print("Testing slise regression")
    X, Y, mod = data_create2(40, 5)
    w = np.random.uniform(size=40) + 0.5
    reg1 = regression(
        X, Y, epsilon=0.1, lambda1=1e-4, lambda2=1e-4, intercept=True, normalise=True
    )
    Yp = mat_mul_inter(X, reg1.coefficients)
    # Yn = reg1._scale.scale_y(Y)
    Ynp = mat_mul_inter(reg1._scale.scale_x(X), reg1._alpha)
    Ypn = reg1._scale.scale_y(Yp)
    # S = (Y - Yp) ** 2 < reg1.epsilon ** 2
    # Sn = (Yn - Ynp) ** 2 < reg1.epsilon_orig ** 2
    assert np.allclose(
        Ypn, Ynp
    ), f"The predicted Y's are not the same {np.max(np.abs(Ynp - Ypn))}"
    assert reg1.score() <= 0, f"SLISE loss should be negative ({reg1.score()})"
    assert 1.0 >= reg1.subset().mean() > 0.7
    reg2 = regression(
        X,
        Y,
        epsilon=0.1,
        lambda1=1e-4,
        lambda2=1e-4,
        intercept=True,
        normalise=False,
    )
    assert reg2.score() <= 0, f"SLISE loss should be negative ({reg2.score()})"
    assert 1.0 >= reg2.subset().mean() > 0.5
    reg3 = regression(
        X,
        Y,
        epsilon=0.1,
        lambda1=0,
        lambda2=0,
        intercept=True,
        normalise=False,
    )
    assert reg3.score() <= 0, f"SLISE loss should be negative ({reg3.score()})"
    assert 1.0 >= reg3.subset().mean() > 0.5
    reg4 = regression(
        X,
        Y,
        epsilon=0.1,
        lambda1=1e-4,
        lambda2=1e-4,
        intercept=True,
        normalise=False,
        weight=w,
    )
    assert reg4.score() <= 0, f"SLISE loss should be negative ({reg4.score()})"
    assert 1.0 >= reg4.subset().mean() > 0.4


def test_slise_exp():
    print("Testing slise explanation")
    np.random.seed(49)
    X, Y, mod = data_create2(100, 5)
    Y2 = sigmoid(Y)
    w = np.random.uniform(size=100) + 0.5
    x = np.random.normal(size=5)
    y = np.random.normal()
    reg = explain(X, Y, 0.1, x, y, lambda1=1e-4, lambda2=1e-4, normalise=True)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert y == approx(reg.predict(x))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, 17, lambda1=0.01, lambda2=0.01, normalise=True)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert Y[17] == approx(reg.predict(X[17]))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, x, y, lambda1=0.01, lambda2=0.01, normalise=False)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert y == approx(reg.predict(x))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, x, y, lambda1=0, lambda2=0, normalise=False)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert y == approx(reg.predict(x))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, 18, lambda1=0.01, lambda2=0.01, normalise=False)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert Y[18] == approx(reg.predict(X[18]))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, 19, lambda1=0, lambda2=0, normalise=False)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert Y[19] == approx(reg.predict(X[19]))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y, 0.1, 19, lambda1=0.01, lambda2=0.01, weight=w, normalise=False)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert Y[19] == approx(reg.predict(X[19]))
    assert 1.0 >= reg.subset().mean() > 0.0
    reg = explain(X, Y2, 0.5, 20, weight=w, normalise=True, logit=True)
    assert reg.score() <= 0, f"Slise loss should usually be <=0 ({reg.score():.2f})"
    assert Y2[20] == approx(reg.predict(X[20]))
    assert 1.0 >= reg.subset().mean() > 0.0


def test_normalised():
    np.random.seed(49)
    X, Y, mod = data_create2(40, 5)
    x = np.random.normal(size=5)
    y = np.random.normal()
    reg = explain(X, Y, 0.1, x, y, lambda1=1e-4, lambda2=1e-4, normalise=True)
    assert reg.coefficients.shape == reg.normalised(True).shape
    assert reg.coefficients.shape[0] > reg.normalised(False).shape[0]
    assert np.allclose(
        reg.coefficients, reg._scale.unscale_model(reg.normalised(False))
    )
    threads = numba.get_num_threads()
    regression(
        X,
        Y,
        epsilon=0.1,
        lambda1=1e-4,
        lambda2=1e-4,
        intercept=True,
        normalise=True,
        num_threads=1,
    )
    assert (
        threads == numba.get_num_threads()
    ), "Numba threads not reset correctly after optimisation"
    assert reg.coefficients.shape == reg.normalised(True).shape
    assert reg.coefficients.shape[0] > reg.normalised(False).shape[0]
    assert np.allclose(
        reg.coefficients, reg._scale.unscale_model(reg.normalised(False))
    )



"""
SLISE - Sparse Linear Subset Explanations
-----------------------------------------

The SLISE algorithm can be used for both robust regression and to explain outcomes from black box models.
See [slise.slise.regression][] and [slise.slise.explain][] for referense.


In robust regression we fit regression models that can handle data that
contains outliers. SLISE accomplishes this by fitting a model such that
the largest possible subset of the data items have an error less than a
given value. All items with an error larger than that are considered
potential outliers and do not affect the resulting model.

SLISE can also be used to provide local model-agnostic explanations for
outcomes from black box models. To do this we replace the ground truth
response vector with the predictions from the complex model. Furthermore, we
force the model to fit a selected item (making the explanation local). This
gives us a local approximation of the complex model with a simpler linear
model. In contrast to other methods SLISE creates explanations using real
data (not some discretised and randomly sampled data) so we can be sure that
all inputs are valid (i.e. in the correct data manifold, and follows the
constraints used to generate the data, e.g., the laws of physics).


More in-depth details about the algorithm can be found in the papers:

Björklund A., Henelius A., Oikarinen E., Kallonen K., Puolamäki K.
Sparse Robust Regression for Explaining Classifiers.
Discovery Science (DS 2019).
Lecture Notes in Computer Science, vol 11828, Springer.
https://doi.org/10.1007/978-3-030-33778-0_27

Björklund A., Henelius A., Oikarinen E., Kallonen K., Puolamäki K.
Robust regression via error tolerance.
Data Mining and Knowledge Discovery (2022).
https://doi.org/10.1007/s10618-022-00819-2

Björklund A., Henelius A., Oikarinen E., Kallonen K., Puolamäki K.
Explaining any black box model using real data.
Frontiers in Computer Science 5:1143904 (2023).
https://doi.org/10.3389/fcomp.2023.1143904
"""

from slise.slise import (  # noqa: F401
    SliseRegression,
    regression,
    SliseExplainer,
    explain,
    SliseWarning,
)
from slise.utils import limited_logit as logit  # noqa: F401
from slise.data import normalise_robust  # noqa: F401



"""
This script contains functions for modifying data, mainly normalisation and PCA.
"""

from typing import NamedTuple, Tuple, Union, Optional

import numpy as np


def add_intercept_column(X: np.ndarray) -> np.ndarray:
    """Add a constant column of ones to the matrix.

    Args:
        X (np.ndarray): Matrix or vector.

    Returns:
        np.ndarray: Matrix / vector where the first column / value is one.
    """
    if len(X.shape) == 1:
        return np.concatenate(([1.0], X))
    return np.concatenate((np.ones((X.shape[0], 1)), X), 1)


def remove_intercept_column(X: np.ndarray) -> np.ndarray:
    """Remove the first column.
    Used to revert [slise.data.add_intercept_column][].

    Args:
        X (np.ndarray): Matrix or vector.

    Returns:
        np.ndarray: Matrix / vector without the first column / value.
    """
    if len(X.shape) == 1:
        return X[1:]
    return X[:, 1:]


def remove_constant_columns(
    X: np.ndarray, epsilon: Optional[float] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Remove columns that are constant from a matrix.
    Used to revert [slise.data.add_constant_columns][].

    Args:
        X (np.ndarray): Data matrix.
        epsilon (Optional[float], optional): Treshold for constant (std < epsilon). Defaults to machine epsilon.

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple of the reduced matrix and a mask showing which columns where retained.
    """
    if epsilon is None:
        epsilon = np.finfo(X.dtype).eps
    std = np.std(X, 0)
    mask = std > epsilon
    return X[:, mask], mask


def add_constant_columns(
    X: np.ndarray, mask: Optional[np.ndarray], intercept: bool = False
) -> np.ndarray:
    """Add (back) contant columns to a matrix.

    Args:
        X (np.ndarray): Data matrix.
        mask (Optional[np.ndarray]): A boolean array showing which columns are already in the matrix.
        intercept (bool, optional): Does X has an intercept (added to it after constant columns where removed). Defaults to False.

    Returns:
        np.ndarray: A matrix with new columns filled with zeros.
    """
    if mask is None:
        return X
    if intercept:
        mask = np.concatenate(([True], mask))
    if len(X.shape) < 2:
        X2 = np.zeros(len(mask), X.dtype)
        X2[mask] = X
        return X2
    else:
        X2 = np.zeros((X.shape[0], len(mask)), X.dtype)
        X2[:, mask] = X
        return X2


def normalise_robust(
    x: np.ndarray, epsilon: Optional[float] = None
) -> Tuple[np.ndarray, Union[float, np.ndarray], Union[float, np.ndarray]]:
    """A robust version of normalisation that uses median and mad (median absolute deviation).
        Any zeros in the scale are replaced by ones to avoid division by zero.

    Args:
        x (np.ndarray): Vector or tensor to normalise.
        epsilon (Optional[float], optional): Threshold for the scale being zero. Defaults to machine epsilon.

    Returns:
        Tuple[np.ndarray, Union[float, np.ndarray], Union[float, np.ndarray]]: Tuple of normalised x, center and scale.
    """
    if epsilon is None:
        epsilon = np.finfo(x.dtype).eps
    if len(x.shape) < 2:
        center = np.median(x)
        x = x - center
        scale = np.median(np.abs(x))
        if scale <= epsilon:
            scale = 1.0
        return x / scale, center, scale
    else:
        center = np.median(x, 0)
        x = x - center[None, :]
        scale = np.median(np.abs(x), 0)
        scale[scale <= epsilon] = 1.0
        return x / scale[None, :], center, scale


def scale_same(
    x: Union[np.ndarray, float],
    center: Union[float, np.ndarray],
    scale: Union[float, np.ndarray],
    constant_colums: Optional[np.ndarray] = None,
    remove_columns: bool = True,
) -> np.ndarray:
    """Scale a matrix or vector the same way as another.

    Args:
        x (np.ndarray): Matrix or vector to scale.
        center (Union[float, np.ndarray]): The center used for the previous scaling.
        scale (Union[float, np.ndarray]): The scale used for the previous scaling.
        constant_colums (Optional[np.ndarray], optional): Boolean mask of constant columns. Defaults to None.
        remove_columns (bool, optional): Should constant columns be removed. Defaults to True.

    Returns:
        np.ndarray: The scaled matrix/vector.
    """
    if isinstance(x, float) or len(x.shape) < 2:
        if constant_colums is not None:
            if not remove_columns:
                y = np.zeros_like(x)
                y[constant_colums] = (x[constant_colums] - center) / scale
                return y
            x = x[constant_colums]
        return (x - center) / scale
    else:
        if constant_colums is not None:
            if not remove_columns:
                y = np.zeros_like(x)
                y[:, constant_colums] = (
                    x[:, constant_colums] - center[None, :]
                ) / scale[None, :]
                return y
            x = x[:, constant_colums]
        return (x - center[None, :]) / scale[None, :]


def unscale_model(
    model: np.ndarray,
    x_center: np.ndarray,
    x_scale: np.ndarray,
    y_center: float = 0.0,
    y_scale: float = 1.0,
    columns: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Scale a linear model such that it matches unnormalised data.

    Args:
        model (np.ndarray): The model for normalised data.
        x_center (np.ndarray): The center used for normalising X.
        x_scale (np.ndarray): The scale used for normalising X.
        y_center (float, optional): The scale used for normalising y. Defaults to 0.0.
        y_scale (float, optional): The center used for normalising y. Defaults to 1.0.
        columns (Optional[np.ndarray], optional): Mask of removed columns (see remove_constant_columns). Defaults to None.

    Returns:
        np.ndarray: The unscaled model.
    """
    if len(model) == len(x_center):
        model = np.concatenate((np.zeros(1, x_center.dtype), model))
    else:
        model = model.copy()
    model[0] = (model[0] - np.sum(model[1:] * x_center / x_scale)) * y_scale + y_center
    model[1:] = model[1:] / x_scale * y_scale
    if columns is not None:
        return add_constant_columns(model, columns, True)
    else:
        return model


class DataScaling(NamedTuple):
    """
    Container class for scaling information
    """

    x_center: np.ndarray
    x_scale: np.ndarray
    y_center: float
    y_scale: float
    columns: np.ndarray

    def scale_x(self, x: np.ndarray, remove_columns: bool = True) -> np.ndarray:
        """Scale a x matrix / vector using the stored scaling information.
        See [slise.data.scale_same][].

        Args:
            x (np.ndarray): New x matrix / vector.
            remove_columns (bool, optional): Remove columns according to the stored information. Defaults to True.

        Returns:
            np.ndarray: Scaled matrix / vector.
        """
        return scale_same(x, self.x_center, self.x_scale, self.columns, remove_columns)

    def scale_y(self, y: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Scale a y vector / scalar using the stored scaling information.
        See [slise.data.scale_same][].

        Args:
            y (np.ndarray): New y vector / scalar.

        Returns:
            np.ndarray: Scaled y vector / scalar.
        """
        return scale_same(y, self.y_center, self.y_scale)

    def unscale_model(self, model: np.ndarray) -> np.ndarray:
        """Unscale a linear model.
        See [slise.data.unscale_model][].

        Args:
            model (np.ndarray): Linear model operating on scaled data.

        Returns:
            np.ndarray: Linear model operating on unscaled data.
        """
        return unscale_model(
            model,
            self.x_center,
            self.x_scale,
            self.y_center,
            self.y_scale,
            self.columns,
        )


def pca_simple(
    x: np.ndarray, dimensions: int = 10, tolerance: float = 1e-10
) -> Tuple[np.ndarray, np.ndarray]:
    """Fit and use PCA for dimensionality reduction.

    Args:
        x (np.ndarray): Matrix to reduce.
        dimensions (int, optional): The number of dimensions to return. Defaults to 10.
        tolerance (float, optional): Threshold for variance being zero. Defaults to 1e-10.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Tuple of the reduced matrix and PCA rotation matrix.
    """
    if len(x.shape) == 1:
        return x, 1.0
    dimensions = min(dimensions, *x.shape)
    u, s, v = np.linalg.svd(x, False, True, False)
    dimensions = max(1, np.sum(s[: min(dimensions, len(s))] > s[0] * tolerance))
    return u[:, :dimensions].dot(np.diag(s[:dimensions])), v[:dimensions, :]


def pca_rotate(x: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Use a trained PCA for dimensionality reduction.
    See [slise.data.pca_simple][] for how to obtain a rotation matrix.

    Args:
        x (np.ndarray): Matrix to reduce.
        v (np.ndarray): PCA rotation matrix.

    Returns:
        np.ndarray: The reduced matrix.
    """
    return x @ v.T


def pca_invert(x: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Revert a PCA dimensionality reduction.
    See [slise.data.pca_simple][] for how to obtain a rotation matrix.

    Args:
        x (np.ndarray): Matrix to expand.
        v (np.ndarray): PCA rotation matrix.

    Returns:
        np.ndarray: The expanded matrix.
    """
    return x @ v


def pca_rotate_model(model: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Transform a linear model to work in PCA reduced space.
    See [slise.data.pca_simple][] for how to obtain a rotation matrix.

    Args:
        model (np.ndarray): Linear model coefficients.
        v (np.ndarray): PCA rotation matrix.

    Returns:
        np.ndarray: The transformed linear model.
    """
    if len(model) > v.shape[1]:
        return np.concatenate((model[:1], v @ model[1:]))
    return v @ model


def pca_invert_model(model: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Transform a linear model from PCA space to "normal" space.
    See [slise.data.pca_simple][] for how to obtain a rotation matrix.

    Args:
        model (np.ndarray): Linear model coefficients (in PCA space).
        v (np.ndarray): PCA rotation matrix.

    Returns:
        np.ndarray: The transformed linear model.
    """
    if len(model) > v.shape[0]:
        return np.concatenate((model[:1], v.T @ model[1:]))
    return v.T @ model



# These tests check if the plotting functions run

import numpy as np
from matplotlib import pyplot as plt
from slise import regression, explain
from slise.plot import get_explanation_order

from .utils import data_create2


def test_plot2d():
    print("Testing 2D plots")
    X, Y, mod = data_create2(40, 1)
    reg = regression(X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=False)
    reg.plot_2d(fig=plt.figure())
    reg = regression(
        X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=True, normalise=True
    )
    reg.plot_2d(fig=plt.figure())
    exp = explain(X, Y, 0.1, 5, lambda1=1e-4, lambda2=1e-4)
    exp.plot_2d(fig=plt.figure())
    Y -= Y.min() - 0.01
    Y /= Y.max() + 0.01
    exp = explain(X, Y, 1.0, 5, lambda1=1e-4, lambda2=1e-4, logit=True)
    exp.plot_2d(fig=plt.figure())
    # plt.show()
    plt.close("all")


def test_print():
    X, Y, mod = data_create2(40, 5)
    reg = regression(X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=True)
    reg.print()
    reg.print(variables=[str(i) for i in range(5)], decimals=2, num_var=4)
    reg = regression(X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=False)
    reg.print()
    reg.print(variables=[str(i) for i in range(5)], decimals=2, num_var=4)
    exp = explain(X, Y, 0.1, 5, lambda1=1e-4, lambda2=1e-4)
    exp.print()
    exp.print(classes=["asd", "bds"], variables=[str(i) for i in range(5)], num_var=4)
    exp.print(classes="bds", decimals=2, num_var=4)
    Y = Y - Y.min() + 0.01
    Y /= Y.max() + 0.01
    exp = explain(X, Y, 1.0, 5, lambda1=1e-4, lambda2=1e-4, normalise=True, logit=True)
    exp.print()
    exp.print(classes=["asd", "bds"], variables=[str(i) for i in range(5)], num_var=4)
    exp.print(classes="bds", decimals=2, num_var=4)


def test_dist():
    try:
        X, Y, mod = data_create2(40, 5)
        reg = regression(X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=False)
        reg.plot_dist(fig=plt.figure())
        reg.plot_dist(variables=[str(i) for i in range(5)], order=3, fig=plt.figure())
        reg.plot_dist(title="asd", order=range(1, 4), decimals=5, fig=plt.figure())
        reg = regression(
            X, Y, 0.1, lambda1=1e-4, lambda2=1e-4, intercept=True, normalise=True
        )
        reg.plot_dist(fig=plt.figure())
        reg.plot_dist(variables=[str(i) for i in range(5)], order=3, fig=plt.figure())
        reg.plot_dist(title="asd", order=range(1, 4), decimals=5, fig=plt.figure())
        exp = explain(X, Y, 0.1, 5, lambda1=1e-4, lambda2=1e-4)
        exp.plot_dist(fig=plt.figure())
        exp.plot_dist(variables=[str(i) for i in range(5)], order=3, fig=plt.figure())
        exp.plot_dist(title="asd", order=range(1, 4), decimals=5, fig=plt.figure())
        Y = Y - Y.min() - 0.01
        Y /= Y.max() + 0.01
        exp = explain(
            X, Y, 1.0, 5, lambda1=1e-4, lambda2=1e-4, normalise=True, logit=True
        )
        exp.plot_dist(fig=plt.figure())
        exp.plot_dist(variables=[str(i) for i in range(5)], order=3, fig=plt.figure())
        exp.plot_dist(title="asd", order=range(1, 4), decimals=5, fig=plt.figure())
        reg.plot_subset(fig=plt.figure())
        exp.plot_subset(fig=plt.figure())
    finally:
        plt.close("all")


def test_img():
    print("Testing image plots")
    X, Y, mod = data_create2(200, 16)
    X[:, 6] = X[:, 9] = X[:, 11] = 0
    exp = explain(X, Y, 0.1, 5, lambda1=1e-4, lambda2=1e-4)
    exp.plot_image(4, 4, fig=plt.figure())
    # plt.show()
    plt.close("all")


def test_order():
    def check(a, b, **kwargs):
        assert np.allclose(a[get_explanation_order(a, **kwargs)], b)

    a = np.arange(5)
    check(a, [0, 4, 3, 2, 1], intercept=True, min=4)
    check(a, [4, 3, 2, 1], intercept=False, min=4)
    check(a, [4, 3, 2], intercept=False, max=3)
    check(a, [0, 4, 3, 2], intercept=True, max=3)
    a = np.arange(-3, 4)
    check(a, [-3, 3, 2, 1, -1, -2], intercept=True, min=4)
    check(a, [3, 2, 1, -1, -2, -3], intercept=False, min=4)
    check(a, [-3, 3, 2, -2], intercept=True, max=3)
    check(a, [3, 2, -2, -3], intercept=False, max=3)



"""
This script contains functions for plotting SLISE solutions.
"""

from collections import OrderedDict
from typing import List, Sequence, Tuple, Union, Optional
from warnings import warn

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, Normalize
from matplotlib.patches import Patch
from matplotlib.pyplot import Figure
from scipy.special import expit as sigmoid
from scipy.stats import gaussian_kde

from slise.utils import SliseException, SliseWarning, mat_mul_inter

# SLISE colors, for unified identity
SLISE_ORANGE = "#fda411"
SLISE_PURPLE = "#9887cb"
SLISE_DARKORANGE = "#e66101"
SLISE_DARKPURPLE = "#5e3c99"
SLISE_COLORMAP = LinearSegmentedColormap.from_list(
    "SLISE", [SLISE_DARKORANGE, SLISE_ORANGE, "#ffffff", SLISE_PURPLE, SLISE_DARKPURPLE]
)
BW_COLORMAP = LinearSegmentedColormap.from_list("BW", ["black", "white"])


def fill_column_names(
    names: Optional[List[str]] = None, amount: int = -1, intercept: bool = False
) -> List[str]:
    """Make sure the list of column names is of the correct size.

    Args:
        names (Optional[List[str]], optional): Prefilled list of column names. Defaults to None.
        amount (int, optional): Number of columns. Defaults to -1.
        intercept (bool, optional): Should an intercept column be added. Defaults to False.

    Returns:
        List[str]: List of column names.
    """
    if amount < 1:
        return names
    if names is None:
        if intercept:
            return ["Intercept"] + ["Variable %d" % i for i in range(amount)]
        else:
            return ["Variable %d" % i for i in range(amount)]
    if len(names) > amount:
        warn("Too many column names given", SliseWarning)
        names = names[:amount]
    if len(names) < amount:
        warn("Too few column names given", SliseWarning)
        names = names + ["Variable %d" % i for i in range(len(names), amount)]
    if intercept:
        return ["Intercept"] + names
    else:
        return names


def fill_prediction_str(
    y: float,
    Y: Optional[np.ndarray] = None,
    classes: Union[List[str], str, None] = None,
    decimals: int = 3,
) -> str:
    """Create a string describing the prediction

    Args:
        y (float): The prediction.
        Y (Optional[np.ndarray]): Vector of predictions (used to guess if the predictions are probabilities). Defaults to None.
        classes (Union[List[str], str, None], optional): List of class names (starting with the negative class), or singular class name. Defaults to None.
        decimals (int, optional): How many decimals hsould be written. Defaults to 3.

    Returns:
        str: Description of the prediction.
    """
    if classes is not None:
        prob = Y is not None and (0 <= Y.min() < 0.5) and (0.5 < Y.max() <= 1)
        if isinstance(classes, str):
            if prob:
                return f"Predicted: {y*100:.{decimals}f}% {classes[0]}"
            else:
                return f"Predicted: {y:.{decimals}f} {classes}"
        else:
            if prob:
                if y > 0.5:
                    return f"Predicted: {y*100:.{decimals}f}% {classes[1]}"
                else:
                    return f"Predicted: {(1-y)*100:.{decimals}f}% {classes[0]}"
            else:
                if y > 0:
                    return f"Predicted: {y:.{decimals}f} {classes[1]}"
                else:
                    return f"Predicted: {-y:.{decimals}f} {classes[0]}"
    else:
        return f"Predicted: {y:.{decimals}f}"


def extended_limits(
    x: np.ndarray, extension: float = 0.05, steps: int = 2
) -> np.ndarray:
    """Create limits that extend a fraction larger than the data.

    Args:
        x (np.ndarray): The data.
        extension (float, optional): How much should the limits extend. Defaults to 0.05.
        steps (int, optional): Number of points in the limit. Defaults to 2.

    Returns:
        np.ndarray: The limit as a vector of points.
    """
    min = np.min(x)
    max = np.max(x)
    diff = max - min
    if steps <= 2:
        return np.array([min - diff * extension, max + diff * extension])
    else:
        return np.linspace(min - diff * extension, max + diff * extension, steps)


def get_explanation_order(
    alpha: np.ndarray,
    intercept: bool = True,
    min: int = 5,
    max: int = -1,
    th: float = 1e-6,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the order in which to show the variables in the plots.

    Args:
        alpha (np.ndarray): Linear model.
        intercept (bool, optional): Does the model include an intercept. Defaults to True.
        min (int, optional): If the number of variables is larger than this, hide the zeroes. Defaults to 5.
        max (int, optional): If `max > 0`, select the top variables. Defaults to -1.
        th ([type], optional): Threshold for zero. Defaults to 1e-6.

    Returns:
        Tuple[np.ndarray, np.ndarray]: The order of the variables in the explanation
    """
    if intercept:
        order = np.argsort(alpha[1:]) + 1
    else:
        order = np.argsort(alpha)
    if len(order) > min:
        order = order[np.nonzero(alpha[order])]
        if len(order) > min:
            order = order[np.abs(alpha[order]) > np.max(np.abs(alpha)) * th]
    if max > 0 and len(order) > max:
        nth = -np.partition(-np.abs(alpha), max - 1)[max - 1]
        order = order[np.abs(alpha[order]) >= nth]
    if intercept:
        order = np.concatenate((order, np.zeros(1, order.dtype)))
    return np.flip(order)


def print_slise(
    coefficients: np.ndarray,
    intercept: bool,
    subset: np.ndarray,
    loss: float,
    epsilon: float,
    variables: Optional[List[str]] = None,
    title: str = "SLISE",
    decimals: int = 3,
    num_var: int = 10,
    unscaled: Optional[np.ndarray] = None,
    unscaled_y: Union[None, float] = None,
    terms: Optional[np.ndarray] = None,
    scaled: Optional[np.ndarray] = None,
    alpha: Optional[np.ndarray] = None,
    scaled_terms: Optional[np.ndarray] = None,
    classes: Optional[List[str]] = None,
    unscaled_preds: Optional[np.ndarray] = None,
    logit: bool = False,
):
    """Print the results from SLISE.

    Args:
        coefficients (np.ndarray): The linear model coefficients.
        intercept (bool): Is the first coefficient an intercept.
        subset (np.ndarray): Subset mask.
        loss (float): SLISE loss.
        epsilon (float): (Unscaled) error tolerance.
        variables (Optional[List[str]], optional): Variable names. Defaults to None.
        title (str, optional): Title to print first. Defaults to "SLISE".
        decimals (int, optional): Number of decimals to print. Defaults to 3.
        num_var (int, optional): Exclude zero weights if there are too many variables. Defaults to 10.
        unscaled (Optional[np.ndarray], optional): Unscaled x (explained item). Defaults to None.
        unscaled_y (Union[None, float], optional): Unscaled y (explained outcome). Defaults to None.
        terms (Optional[np.ndarray], optional): Unscaled terms (coefficients * x). Defaults to None.
        scaled (Optional[np.ndarray], optional): Scaled x (explained item). Defaults to None.
        alpha (Optional[np.ndarray], optional): Scaled model. Defaults to None.
        scaled_terms (Optional[np.ndarray], optional): Scaled terms (alpha * scaled_x). Defaults to None.
        classes (Optional[List[str]], optional): Class names (if applicable). Defaults to None.
        unscaled_preds (Optional[np.ndarray], optional): Unscaled resonse (Y-vector). Defaults to None.
        logit (bool, optional): A logit transformation has been applied. Defaults to False.
    """
    rows = OrderedDict()
    rows["Variable Names:    "] = fill_column_names(
        variables, len(coefficients) - intercept, intercept
    )
    if unscaled is not None:
        rows["Explained Item:"] = [""] + ["%%.%df" % decimals % a for a in unscaled]
        rows["Model Weights:"] = ["%%.%df" % decimals % a for a in coefficients]
    else:
        rows["Coefficients:"] = ["%%.%df" % decimals % a for a in coefficients]
    if terms is not None:
        rows["Prediction Term:"] = ["%%.%df" % decimals % a for a in terms]
    if scaled is not None:
        rows["Normalised Item:"] = [""] + ["%%.%df" % decimals % a for a in scaled]
    if alpha is not None:
        rows["Normalised Weights:"] = ["%%.%df" % decimals % a for a in alpha]
    if scaled_terms is not None:
        rows["Normalised Term:"] = ["%%.%df" % decimals % a for a in scaled_terms]
    col_len = [
        max(8, *vs) + 1
        for vs in zip(*(tuple(len(v) for v in vs) for vs in rows.values()))
    ]
    if len(coefficients) > num_var:
        col_len = [cl if c != 0 else 0 for cl, c in zip(col_len, coefficients)]
    lab_len = max(len(r) for r in rows)
    if title:
        print(title)
    if unscaled_y is not None:
        print(fill_prediction_str(unscaled_y, unscaled_preds, classes, decimals))
    for k in rows:
        print(
            f"{k:<{lab_len}}",
            " ".join([f"{s:>{c}}" for s, c in zip(rows[k], col_len) if c > 0]),
        )
    loss = f"{loss:.{decimals}f}"
    epsilon = f"{epsilon:.{decimals}f}"
    subsize = f"{subset.mean():.{decimals}f}"
    col_len = max(len(loss), len(epsilon), len(subsize), 8)
    print(f"Loss:          {loss   :>{col_len}}")
    print(f"Subset:        {subsize:>{col_len}}")
    print(f"Epsilon:       {epsilon:>{col_len}}")
    if logit and unscaled_preds is not None:
        if isinstance(classes, list) and len(classes) == 2:
            print(
                f"Class Balance: {(unscaled_preds[subset] > 0.5).mean() * 100:>.{decimals}f}% {classes[0]} | {(unscaled_preds[subset] < 0.5).mean() * 100:>.{decimals}f}% {classes[1]}"
            )
        else:
            print(
                f"Class Balance: {(unscaled_preds[subset] > 0.5).mean() * 100:>.{decimals}f}% | {(unscaled_preds[subset] < 0.5).mean() * 100:>.{decimals}f}%"
            )


def plot_2d(
    X: np.ndarray,
    Y: np.ndarray,
    model: np.ndarray,
    epsilon: float,
    x: Optional[np.ndarray] = None,
    y: Optional[float] = None,
    logit: bool = False,
    title: str = "SLISE for Robust Regression",
    label_x: str = "x",
    label_y: str = "y",
    decimals: int = 3,
    fig: Optional[Figure] = None,
):
    """Plot the regression/explanation in a 2D scatter plot with a line for the regression model (and the explained item marked).

    Args:
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        model (np.ndarray): Linear model.
        epsilon (float): Error tolerance.
        x (Optional[np.ndarray], optional): Explained item. Defaults to None.
        y (Optional[float], optional): Explained outcome. Defaults to None.
        logit (bool, optional): Should Y be logit-transformed. Defaults to False.
        title (str, optional): Plot title. Defaults to "SLISE for Robust Regression".
        label_x (str, optional): X-axis label. Defaults to "x".
        label_y (str, optional): Y-axis label. Defaults to "y".
        decimals (int, optional): Number of decimals when writing numbers. Defaults to 3.
        fig (Optional[Figure], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.

    Raises:
        SliseException: If the data has too many dimensions.
    """
    if fig is None:
        plot = True
        fig, ax = plt.subplots()
    else:
        ax = fig.subplots()
        plot = False
    if X.size != Y.size:
        raise SliseException(f"Can only plot 1D data, |Y| = {Y.size} != {X.size} = |X|")
    x_limits = extended_limits(X, 0.03, 20 if logit else 2)
    y_limits = mat_mul_inter(x_limits[:, None], model)
    if logit:
        ax.fill_between(
            x_limits,
            sigmoid(y_limits + epsilon),
            sigmoid(y_limits - epsilon),
            color=SLISE_PURPLE + "33",
            label="Subset",
        )
        y_limits = sigmoid(y_limits)
    else:
        ax.fill_between(
            x_limits,
            y_limits + epsilon,
            y_limits - epsilon,
            color=SLISE_PURPLE + "33",
            label="Subset",
        )
    ax.plot(X.ravel(), Y, "o", color="black", label="Dataset")
    if x is not None and y is not None:
        ax.plot(x_limits, y_limits, "-", color=SLISE_PURPLE, label="Model")
        ax.plot(x, y, "o", color=SLISE_ORANGE, label="Explained Item")
    else:
        ax.plot(x_limits, y_limits, "-", color=SLISE_ORANGE, label="Model")
    formula = ""
    if isinstance(model, float):
        formula = f"{model:.{decimals}f} * {label_x}"
    elif len(model.flat) == 1:
        formula = f"{model.flat[0]:.{decimals}f} * {label_x}"
    elif np.abs(model[0]) > 1e-8:
        sign = "-" if model[1] < 0.0 else "+"
        formula = f"{model[0]:.{decimals}f} {sign} {abs(model[1]):.{decimals}f} $\\cdot$ {label_x}"
    else:
        formula = f"{model[1]:.{decimals}f} * {label_x}"
    if logit:
        formula = f"$\\sigma$({formula})"
    ax.legend()
    ax.set_xlabel(label_x)
    ax.set_ylabel(label_y)
    ax.set_title(f"{title}: {label_y} = {formula}")
    fig.tight_layout()
    if plot:
        plt.show()


def plot_dist(
    X: np.ndarray,
    Y: np.ndarray,
    model: np.ndarray,
    subset: np.ndarray,
    alpha: Optional[np.ndarray] = None,
    x: Optional[np.ndarray] = None,
    y: Optional[float] = None,
    terms: Optional[np.ndarray] = None,
    norm_terms: Optional[np.ndarray] = None,
    title: str = "SLISE Explanation",
    variables: Optional[List[str]] = None,
    order: Union[None, int, Sequence[int]] = None,
    decimals: int = 3,
    fig: Optional[Figure] = None,
):
    """Plot the SLISE result with density distributions for the dataset and barplot for the model.

    Args:
        X (np.ndarray): Data matrix.
        Y (np.ndarray): Response vector.
        model (np.ndarray): Linear model.
        subset (np.ndarray): Selected subset.
        alpha (Optional[np.ndarray]): Scaled model. Defaults to None.
        x (Optional[np.ndarray], optional): The explained item (if it is an explanation). Defaults to None.
        y (Optional[float], optional): The explained outcome (if it is an explanation). Defaults to None.
        terms (Optional[np.ndarray], optional): Term vector (unscaled x*alpha), if available. Defaults to None.
        norm_terms (Optional[np.ndarray], optional): Term vector (scaled x*alpha), if available. Defaults to None.
        title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
        order (Union[None, int, Sequence[int]], optional): Select variables (None: all, int: largest, selected). Defaults to all.
        variables (Optional[List[str]], optional): Names for the (columns/) variables. Defaults to None.
        decimals (int, optional): Number of decimals when writing numbers. Defaults to 3.
        fig (Optional[Figure], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
    """
    # Values and order
    variables = fill_column_names(variables, X.shape[1], True)
    if alpha is None:
        noalpha = True
        alpha = model
    else:
        noalpha = False
    order_offset = 0
    if len(model) == X.shape[1]:
        model = np.concatenate((np.zeros(1, model.dtype), model))
        alpha = np.concatenate((np.zeros(1, model.dtype), alpha))
        order_offset = 1
        variables[0] = ""
    if order is None:
        order = get_explanation_order(np.abs(alpha), True)
    elif isinstance(order, int):
        order = get_explanation_order(np.abs(alpha), True, max=order)
    else:
        order = [0] + [i + order_offset for i in order if i + order_offset != 0]
    model = model[order]
    alpha = alpha[order]
    if terms is not None:
        terms = terms[order]
    if norm_terms is not None:
        norm_terms = norm_terms[order]
    variables = [variables[i] for i in order]
    subsize = subset.mean()

    # Figures:
    if isinstance(fig, Figure):
        plot = False
        axs = fig.subplots(len(order), 2, squeeze=False)
    else:
        plot = True
        fig, axs = plt.subplots(len(order), 2, squeeze=False)
    fig.suptitle(title)

    # Density plots

    def fill_density(ax, X, x, n):
        if np.var(X) == 0:
            X = np.random.normal(X[0], 1e-8, len(X))
        kde1 = gaussian_kde(X, 0.2)
        if np.sum(subset) > 1:
            kde2 = gaussian_kde(X[subset], 0.2)
        else:
            kde2 = lambda x: x * 0  # noqa: E731
        lim = extended_limits(X, 0.1, 100)
        ax.plot(lim, kde1(lim), color="black", label="Dataset")
        ax.plot(
            lim,
            kde2(lim) * subsize,
            color=SLISE_PURPLE,
            label=f"Subset: {subsize * 100:.0f}%",
        )
        if x is not None:
            ax.relim()
            ax.vlines(x, *ax.get_ylim(), color=SLISE_ORANGE, label="Explained Item")
        ax.set_yticks([])
        ax.set_ylabel(
            n, rotation=0, horizontalalignment="right", verticalalignment="center"
        )

    if x is None and y is None:
        fill_density(axs[0, 0], Y, y, "Response")
    else:
        fill_density(axs[0, 0], Y, y, "Prediction")
    axs[0, 0].legend()
    axs[0, 0].set_title("Dataset Distribution")
    for i, k, n in zip(range(1, len(order)), order[1:], variables[1:]):
        fill_density(axs[i, 0], X[:, k - 1], x[k - 1] if x is not None else None, n)

    # Bar plots
    def text(x, y, v):
        if v != 0:
            axbig.text(
                x,
                y,
                f"{v:.{decimals}f}",
                ha="center",
                va="center",
                bbox=dict(boxstyle="round", fc="white", ec="grey", alpha=0.75),
            )

    gs = axs[0, 1].get_gridspec()
    for ax in axs[:, 1]:
        ax.remove()
    axbig = fig.add_subplot(gs[:, 1])
    if x is None or y is None:
        axbig.set_title("Linear Model")
    else:
        axbig.set_title("Explanation")
    ticks = np.arange(len(variables))
    axbig.set_yticks(ticks)
    axbig.set_yticklabels(variables)
    axbig.set_ylim(bottom=ticks[0] - 0.45, top=ticks[-1] + 0.45)
    axbig.invert_yaxis()
    if terms is None and noalpha:
        column_color = [SLISE_ORANGE if v < 0 else SLISE_PURPLE for v in alpha]
        axbig.barh(ticks, alpha, color=column_color)
        for y, v in zip(ticks, model):
            text(0, y, v)
    elif terms is None and not noalpha:
        axbig.barh(
            ticks - 0.2,
            model / np.max(np.abs(model)),
            height=0.35,
            color=SLISE_PURPLE,
            label="Coefficients",
        )
        axbig.barh(
            ticks + 0.2,
            alpha / np.max(np.abs(alpha)),
            height=0.35,
            color=SLISE_ORANGE,
            label="Normalised",
        )
        for y, a, m in zip(ticks, alpha, model):
            text(0, y, m)
            text(0, y, a)
        axbig.set_xticks([])
        axbig.legend()
    elif norm_terms is None:
        axbig.barh(
            ticks[1:] - 0.2,
            model[1:] / np.max(np.abs(model)),
            height=0.35,
            color=SLISE_PURPLE,
            label="Linear Model",
        )
        axbig.barh(
            ticks[0],
            model[0] / np.max(np.abs(model)),
            height=0.35,
            color=SLISE_PURPLE,
        )
        axbig.barh(
            ticks[1:] + 0.2,
            terms[1:] / np.max(np.abs(terms[1:])),
            height=0.35,
            color=SLISE_ORANGE,
            label="Prediction Term",
        )
        for y, a, m in zip(ticks, terms, model):
            if y == ticks[0]:
                text(0, y, m)
                continue
            text(0, y - 0.2, m)
            text(0, y + 0.2, a)
        axbig.set_xticks([])
        axbig.legend()
    else:
        axbig.barh(
            ticks[1:] - 0.33,
            model[1:] / np.max(np.abs(model)),
            height=0.2,
            color=SLISE_PURPLE,
            label="Linear Model",
        )
        axbig.barh(
            ticks[0] - 0.11,
            model[0] / np.max(np.abs(model)),
            height=0.2,
            color=SLISE_PURPLE,
        )
        axbig.barh(
            ticks[1:] - 0.11,
            alpha[1:] / np.max(np.abs(alpha)),
            height=0.2,
            color=SLISE_DARKPURPLE,
            label="Normalised Model",
        )
        axbig.barh(
            ticks[0] + 0.11,
            alpha[0] / np.max(np.abs(alpha)),
            height=0.2,
            color=SLISE_DARKPURPLE,
        )
        axbig.barh(
            ticks[1:] + 0.11,
            terms[1:] / np.max(np.abs(terms[1:])),
            height=0.2,
            color=SLISE_ORANGE,
            label="Prediction Term",
        )
        axbig.barh(
            ticks[1:] + 0.33,
            norm_terms[1:] / np.max(np.abs(norm_terms[1:])),
            height=0.2,
            color=SLISE_DARKORANGE,
            label="Normalised Term",
        )
        for y, i1, i2, m1, m2 in zip(ticks, terms, norm_terms, model, alpha):
            if y == ticks[0]:
                text(0, y - 0.11, m1)
                text(0, y + 0.11, m2)
                continue
            text(0, y - 0.33, m1)
            text(0, y - 0.11, m2)
            text(0, y + 0.11, i1)
            text(0, y + 0.33, i2)
        axbig.set_xticks([])
        axbig.legend()
    axbig.yaxis.tick_right()

    # Meta:
    fig.tight_layout()
    if plot:
        plt.show()


def plot_image(
    x: np.ndarray,
    y: float,
    Y: np.ndarray,
    model: np.ndarray,
    width: int,
    height: int,
    saturated: bool = True,
    title: str = "SLISE Explanation",
    classes: Union[List, str, None] = None,
    decimals: int = 3,
    fig: Optional[Figure] = None,
):
    """Plot an explanation for a black and white image (e.g. MNIST).

    Args:
        x (np.ndarray): The explained item.
        y (float): The explained outcome.
        Y (np.ndarray): Dataset response vector (used for guessing prediction formatting).
        model (np.ndarray): The approximating model.
        width (int): The width of the image.
        height (int): The height of the image.
        saturated (bool, optional): Should the explanation be more saturated. Defaults to True.
        title (str, optional): Title of the plot. Defaults to "SLISE Explanation".
        classes (Union[List, str, None], optional): List of class names (first the negative, then the positive), or a single (positive) class name. Defaults to None.
        decimals (int, optional): The number of decimals to write. Defaults to 3.
        fig (Optional[Figure], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
    """
    # intercept = model[0]
    model = model[1:]
    model.shape = (width, height)
    x.shape = (width, height)
    if saturated:
        model = sigmoid(model * (4 / np.max(np.abs(model))))
    if fig is None:
        fig, [ax1, ax2] = plt.subplots(1, 2)
        plot = True
    else:
        [ax1, ax2] = fig.subplots(1, 2)
        plot = False
    fig.suptitle(title)
    # Image
    ax1.imshow(x, cmap=BW_COLORMAP)
    ax1.set_xticks([])
    ax1.set_yticks([])
    ax1.set_title("Explained Item")
    ax1.set_xlabel(fill_prediction_str(y, Y, classes, decimals))
    # Explanation Image
    ax2.imshow(
        model,
        interpolation="none",
        cmap=SLISE_COLORMAP,
        norm=Normalize(vmin=-0.1, vmax=1.1),
    )
    ax2.contour(range(height), range(width), x, levels=1, colors="#00000033")
    ax2.set_xticks([])
    ax2.set_yticks([])
    ax2.set_title("Explanation")
    if classes is None:
        classes = ["Negative", "Positive"]
    elif isinstance(classes, str):
        classes = ["Not " + classes, classes]
    ax2.legend(
        (Patch(facecolor=SLISE_ORANGE), Patch(facecolor=SLISE_PURPLE)),
        classes[:2],
        loc="upper center",
        bbox_to_anchor=(0.5, -0.01),
        ncol=2,
    )
    fig.tight_layout()
    if plot:
        plt.show()


def plot_dist_single(
    data: np.ndarray,
    subset: np.ndarray,
    item: Optional[float] = None,
    title: str = "Response Distribution",
    decimals: int = 0,
    fig: Optional[Figure] = None,
):
    """Plot a density distributions for a single variable of the dataset.

    Args:
        data (np.ndarray): Variable vector.
        subset (np.ndarray): Selected subset.
        item (Optional[np.ndarray], optional): The explained item (if it is an explanation). Defaults to None.
        title (str, optional): Title of the plot. Defaults to "Response Distribution".
        decimals (int, optional): Number of decimals when writing the subset size. Defaults to 0.
        fig (Optional[Figure], optional): Pyplot figure to plot on, if None then a new plot is created and shown. Defaults to None.
    """
    subsize = subset.mean()
    if isinstance(fig, Figure):
        ax = fig.subplots(1, 1)
        plot = False
    else:
        fig, ax = plt.subplots(1, 1)
        plot = True
    ax.set_title(title)
    kde1 = gaussian_kde(data, 0.2)
    kde2 = gaussian_kde(data[subset], 0.2)
    lim = extended_limits(data, 0.1, 100)
    ax.plot(lim, kde1(lim), color="black", label="Dataset")
    ax.plot(
        lim,
        kde2(lim) * subsize,
        color=SLISE_PURPLE,
        label=f"Subset: {subsize * 100:.{decimals}f}%",
    )
    if item is not None:
        ax.relim()
        ax.vlines(item, *ax.get_ylim(), color=SLISE_ORANGE, label="Explained Item")
    ax.set_yticks([])
    ax.legend()
    if plot:
        plt.show()
